#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#if defined STM32F40XX
#include "stm32f4xx.h"
#include "stm324xg_eval.h"
#include "stm32_eval_legacy.h"

#else
#include "stm32f10x.h"
#include "stm32_eval.h"
#endif

#include "wl_fw.h"
#include "wl_sdio.h" /* for owl_sdio_mdelay() prototype */

#include <owl/core/owl_loop.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_wifi.h>
#include <owl/core/owl_ap.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_net.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_defs.h>
#include <owl/drivers/owl_uart.h>

#include <owl/util/ttcp.h>
#include <owl/util/httpd/httpd.h>

#ifdef USE_STM324xG_EVAL
#include "stm324xg_eval_lcd.h"
#elif defined USE_STM32100B_EVAL
 #include "STM32100B_EVAL/stm32100b_eval_lcd.h"
#elif defined USE_STM3210B_EVAL
 #include "STM3210B_EVAL/stm3210b_eval_lcd.h"
#elif defined USE_STM3210E_EVAL
 #include "STM3210E_EVAL/stm3210e_eval_lcd.h" 
#elif defined USE_STM3210C_EVAL
 #include "STM3210C_EVAL/stm3210c_eval_lcd.h"
#elif defined USE_STM32100E_EVAL
 #include "STM32100E_EVAL/stm32100e_eval_lcd.h"
#endif

#define MESSAGE1     "   H&D Wireless     " 
#define MESSAGE2     "   HTTP Server on   " 
#ifdef USE_STM324xG_EVAL
  #define MESSAGE3     "   STM324xG-EVAL    "
#elif defined USE_STM32100B_EVAL
  #define MESSAGE3   "   STM32100B-EVAL   " 
#elif defined (USE_STM3210B_EVAL)
  #define MESSAGE3   "   STM3210B-EVAL    " 
#elif defined (USE_STM3210E_EVAL)
  #define MESSAGE3   "   STM3210E-EVAL    " 
#elif defined (USE_STM3210C_EVAL)
  #define MESSAGE3   "   STM3210C-EVAL    "
#elif defined (USE_STM32100E_EVAL)
  #define MESSAGE3   "   STM32100E-EVAL   "   
#endif

#define TIMER_PERIOD_MS 10


/**
 * Delay executiom for the specified number of ms. This function will be called
 * with delays in the 10-20 ms range during fw download and startup of the
 * Wi-Fi device. This function can be implemented with a simple for-loop if
 * desired (beware of optimization). The timing does not have to be accurate as
 * long as the actual delay becomes at least the specified number of ms.
 *
 * NOTE: This function is part of the wl_sdio.h porting layer but we define it
 *       here in main.c rather than owl_sdio.c make owl_sdio.c only depending on
 *       the STM32 peripheral library.
 *
 * @param ms is the minimal amount of time to wait [ms].
 *
 */
void owl_sdio_mdelay(uint32_t ms)
{
        volatile int a = 0;                     
        unsigned int i;                                  
        for (i = 0; i < ms * 5000; i++)         
                a++;                                
}      


/**
 * Called during wl_fw_download() to provide firmware data. This function will
 * be called until the complete firmware image has been obtained.
 *
 * If firmware is stored in external flash, this function could read and return
 * chunks of data by specifiying a partial length in the len parameter.
 *
 * See wl_api.h for more information.
 *
 */ 

static size_t
fw_read_cb(void *ctx, const uint8_t **buf, size_t offset, size_t len)
{
        (void) ctx; /* unused */

        if (NULL == buf)
                return 0;

        *buf = ((const uint8_t *) fw_buf) + offset;
        if (len > (fw_len - offset))
                return fw_len - offset;

        return len;
}


/**
 * Called by owl-debug functions, such as owl_dbg(), owl_info(),
 * owl_printf() ... defined in owl_env.h
 *
 */
void owl_putc(char c)
{
#ifdef USE_STM324xG_EVAL

	/* write a character to the USART */
        USART_SendData(USART3, (uint8_t) c);

	/* Loop until the end of transmission */
        while (USART_GetFlagStatus(USART3, USART_FLAG_TC) == RESET);

        if (c == '\n') {
                USART_SendData(USART3, (uint8_t) '\r');
                while (USART_GetFlagStatus(USART3, USART_FLAG_TC) == RESET);
        }

#else /* STM32F1 */
       
        USART_SendData(EVAL_COM1, (uint8_t) c);
        while (USART_GetFlagStatus(EVAL_COM1, USART_FLAG_TC) == RESET);

        if (c == '\n') {
                USART_SendData(EVAL_COM1, (uint8_t) '\r');
                while (USART_GetFlagStatus(EVAL_COM1, USART_FLAG_TC) == RESET);
        }
#endif
}


/**
 *
 *
 */
void SysTick_Handler()
{
        owltmr_tick(TIMER_PERIOD_MS);
}



#if defined(APP_MODE_WIFI)

/**
 * Called when wifi link status changes, see owl_wifi.c.
 *
 *
 */
static void
link_status_cb(void *ctx, int status)
{
        struct owlsh_ctx *shell = ctx;
        
        owlsh_printf(shell, "link status: %d\r\n", status);

        if (status) {
                owlnet_linkup();
        } else {
                owlnet_linkdown();
        }
}


/**
 * Called when ip adress is obtained or lost, see owl_net.c
 *
 *
 */
static void
ip_status_cb(struct netif *netif, void *ctx)
{
        struct owlsh_ctx *shell = ctx;
        
        if (netif_is_up(netif)) {
                owlsh_printf(shell, "bound to %d.%d.%d.%d\r\n",
                             ip4_addr1(&netif->ip_addr),
                             ip4_addr2(&netif->ip_addr),
                             ip4_addr3(&netif->ip_addr),
                             ip4_addr4(&netif->ip_addr));
        }
        else {
                owlsh_printf(shell, "ip status down\r\n");
        }
}
#endif


/**
 * sbrk
 *
 * Increase program data space. Malloc and related functions depend on this.
 */
caddr_t _sbrk(int incr)
{

        extern char _ebss; /* Defined by the linker */
        static char *heap_end;
        char *prev_heap_end;

        if (heap_end == 0)
                heap_end = &_ebss;

        prev_heap_end = heap_end;

        char *stack = (char *) __get_MSP();
        if (heap_end + incr >  stack) {
                errno = ENOMEM;
                return  (caddr_t) -1;
        }

        heap_end += incr;
        return (caddr_t) prev_heap_end;
}


#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t* file, uint32_t line)
{ 
        /* User can add his own implementation to report the file name and line
           number, ex: printf("Wrong parameters value: file %s on line %d\r\n",
           file, line) */

        /* Infinite loop */
        while (1);
}
#endif


int
httpd_index_cb(void *ctx, int seq, struct http_request *req,
               struct pbuf *p, char *rspbuf, uint16_t rsplen, void *local)
{
        int pos = 0;
        int remain = rsplen;
        int err;

        switch (seq) {
                
        case 0: {

                if ((err = sniprintf(
                             rspbuf + pos, remain,
                             "HTTP/1.0 200 OK\r\n"
                             "Server: lwIP\r\n"
                             "Content-type: text/html\r\n\r\n"
                             "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 "
                             "Frameset//EN\"\"http://www.w3.org/TR/html4/"
                             "frameset.dtd\">")) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;
                   
                if ((err = sniprintf(
                             rspbuf + pos, remain,
                             "<html>"
                             "<head><title>802.11bg Wi-Fi on STM32</title>"
                             "</head><body bgcolor=white text=black>"
                             "<table width=100%%>"
                             "<tr valign=top><td width=80>"
                             "<a href=index.html>"
                             "<img src=image.gif border=0 alt=logo"
                             "title=logo></a>" 
                             "</td><td width=500>"
                             "<h2>802.11bg Wi-Fi on STM32</h2>"
                             "<p>The web page is served by an STM32, using the "
                             "H&D Wireless&rsquo; unique Wi-Fi solution, "
                             "consisting of the HDG104 WLAN SiP and the "
                             "Software SDK, oWL&copy;. The web server is "
                             "running on top of the lwIP TCP/IP stack."
                             "</p><hr><table>")) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;
                break;

        }
                
        case 1: {
                struct netif *netif = owlnet_netif();
                
                if ((err = sniprintf(
                             rspbuf + pos, remain,
                             "<tr><td>IP Address: </td><td><b>%s</b></td></tr>"
                             "<tr><td>MAC Address: </td><td><b>"
                             "%02x:%02x:%02x:%02x:%02x:%02x</b></td></tr>",
                             ip_ntoa(&netif->ip_addr),
                             netif->hwaddr[0], netif->hwaddr[1],
                             netif->hwaddr[2], netif->hwaddr[3],
                             netif->hwaddr[4], netif->hwaddr[5])) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;

                
                if ((err = sniprintf(
                             rspbuf + pos, remain,
                             "<tr><td><form method=get action=index.html>"
                             "<fieldset style='width: 300;'>"
                             "<legend>LED Control</legend>"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED1>&nbsp&nbsp"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED2>&nbsp&nbsp"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED3>&nbsp&nbsp"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED4>&nbsp&nbsp</fieldset></form>"
                             "</td><td>&nbsp;</td></tr></table>"
                             "</body></html>")) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;

                break;
        }

        default: /* done */
                return 0;
                
        }
        
        return pos;
}


int
httpd_led_cb(void *ctx, int seq, struct http_request *req,
             struct pbuf *p, char *rspbuf, uint16_t rsplen, void *local)
{
        Led_TypeDef led = (Led_TypeDef) ctx;

        if (!seq)
                STM_EVAL_LEDToggle(led);
                
        return httpd_index_cb(ctx, seq, req, p, rspbuf, rsplen, local);
}

int
httpd_logo_cb(void *ctx, int seq, struct http_request *req,
              struct pbuf *p, char *rspbuf, uint16_t rsplen, void *local)
{
        static const char hdwireless_gif[] = {
                /* 
                 * "HTTP/1.0 200 OK\n"
                 * "Server: lwIP\n"
                 * "Content-type: image/gif\n\n",
                 * 
                 */
                0x48, 0x54, 0x54, 0x50, 0x2f, 0x31, 0x2e, 0x30, 
                0x20, 0x32, 0x30, 0x30, 0x20, 0x4f, 0x4b, 0x0d, 
                0x0a, 0x53, 0x65, 0x72, 0x76, 0x65, 0x72, 0x3a, 
                0x20, 0x6c, 0x77, 0x49, 0x50, 0x0d, 0x0a, 0x43, 
                0x6f, 0x6e, 0x74, 0x65, 0x6e, 0x74, 0x2d, 0x74, 
                0x79, 0x70, 0x65, 0x3a, 0x20, 0x69, 0x6d, 0x61, 
                0x67, 0x65, 0x2f, 0x67, 0x69, 0x66, 0x0d, 0x0a, 
                0x0d, 0x0a,

                /* gif data */
                0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x45, 0x01, 
                0x7b, 0x00, 0xa1, 0x00, 0x00, 0xff, 0xff, 0xff, 
                0x00, 0x66, 0x99, 0x00, 0x00, 0x00, 0xff, 0xff, 
                0xff, 0x21, 0xfe, 0x11, 0x43, 0x72, 0x65, 0x61, 
                0x74, 0x65, 0x64, 0x20, 0x77, 0x69, 0x74, 0x68, 
                0x20, 0x47, 0x49, 0x4d, 0x50, 0x00, 0x2c, 0x00, 
                0x00, 0x00, 0x00, 0x45, 0x01, 0x7b, 0x00, 0x00, 
                0x02, 0xfe, 0x84, 0x8f, 0xa9, 0xcb, 0xed, 0x0f, 
                0xa3, 0x9c, 0xb4, 0xda, 0x8b, 0xb3, 0xde, 0xbc, 
                0xfb, 0x0f, 0x86, 0xe2, 0x48, 0x96, 0xe6, 0x89, 
                0xa6, 0xea, 0xca, 0xb6, 0xee, 0x0b, 0xc7, 0xf2, 
                0x4c, 0xd7, 0xf6, 0x8d, 0xe7, 0xfa, 0xce, 0xf7, 
                0xfe, 0x0f, 0x0c, 0x0a, 0x87, 0xc4, 0xa2, 0xf1, 
                0x88, 0x4c, 0x2a, 0x97, 0xcc, 0xa6, 0xf3, 0x09, 
                0x8d, 0x4a, 0xa7, 0xd4, 0xaa, 0xf5, 0x8a, 0x65, 
                0x05, 0xb6, 0xdc, 0xae, 0xf7, 0x0b, 0x0e, 0x8b, 
                0xc7, 0x64, 0x71, 0xf6, 0x0c, 0x2d, 0xab, 0xd7, 
                0xec, 0x36, 0x19, 0x0d, 0x3f, 0xba, 0xe7, 0xf4, 
                0xfa, 0x3b, 0x8e, 0xff, 0xd9, 0xf7, 0x7c, 0x7b, 
                0xfe, 0x9f, 0xd3, 0x27, 0x38, 0xc8, 0x06, 0x68, 
                0x38, 0x43, 0x98, 0xa8, 0x18, 0x76, 0xd8, 0xd8, 
                0xb2, 0x08, 0x19, 0x19, 0xe0, 0x48, 0x79, 0x22, 
                0x79, 0xa9, 0x58, 0xa9, 0x29, 0x82, 0xd9, 0x29, 
                0xb8, 0x09, 0xda, 0xe1, 0x39, 0xca, 0x17, 0x6a, 
                0x8a, 0x41, 0x9a, 0x4a, 0x77, 0xca, 0x3a, 0xa1, 
                0xfa, 0xea, 0xd6, 0x2a, 0xeb, 0x00, 0x5b, 0xab, 
                0x36, 0x8b, 0xab, 0x60, 0xbb, 0x6b, 0x96, 0xeb, 
                0x0b, 0xc0, 0x1b, 0xec, 0xf5, 0xeb, 0x2b, 0x6c, 
                0xbc, 0x45, 0x8c, 0x7b, 0x6c, 0x9c, 0x3c, 0xbb, 
                0x2c, 0xdc, 0x2c, 0xfb, 0x1c, 0x1c, 0xdd, 0x3a, 
                0xcd, 0x5b, 0xcd, 0x7a, 0xbd, 0x9b, 0x7d, 0xba, 
                0x6d, 0xdb, 0x6d, 0xfa, 0x5d, 0x1b, 0x1e, 0x3a, 
                0x0e, 0x5b, 0x0e, 0x7a, 0xfe, 0x9a, 0xbe, 0xb9, 
                0xae, 0xda, 0xae, 0xf9, 0x9e, 0x1a, 0x5f, 0x39, 
                0x4f, 0x5a, 0x4f, 0x79, 0x3f, 0x9a, 0xef, 0xb8, 
                0xef, 0xd9, 0xaf, 0xd1, 0xbf, 0x4e, 0x01, 0x0f, 
                0x0d, 0xc4, 0x54, 0xd0, 0xd0, 0xc1, 0x4b, 0x09, 
                0x01, 0x2d, 0x94, 0xd4, 0xf0, 0xcf, 0xc3, 0x48, 
                0x11, 0xf3, 0x4c, 0x84, 0x54, 0x11, 0xcf, 0x1d, 
                0xfe, 0x5d, 0x60, 0x1a, 0x8c, 0xf1, 0x58, 0xa6, 
                0x42, 0x1b, 0x04, 0x02, 0x4a, 0x9a, 0x14, 0xb0, 
                0xe0, 0xe4, 0xc9, 0x94, 0x2a, 0x51, 0x1e, 0x68, 
                0x29, 0x03, 0x66, 0xb8, 0x8d, 0x09, 0x18, 0x31, 
                0xb0, 0xf9, 0x80, 0xa6, 0xab, 0x5b, 0x24, 0x65, 
                0xf6, 0xf4, 0xf9, 0x12, 0x28, 0x00, 0xa1, 0x2d, 
                0x88, 0x36, 0xd3, 0x79, 0x00, 0x27, 0xc7, 0x61, 
                0x12, 0x3e, 0x5e, 0xe0, 0x19, 0x54, 0xa5, 0x82, 
                0x96, 0x25, 0xa7, 0x0a, 0x35, 0xba, 0x02, 0x2b, 
                0x31, 0xa4, 0x06, 0x94, 0xd6, 0x64, 0xda, 0xb4, 
                0xd7, 0x53, 0x9d, 0x54, 0x13, 0x50, 0x35, 0x69, 
                0xf6, 0xaa, 0xd4, 0x18, 0x5a, 0x7f, 0x71, 0x05, 
                0xd6, 0xf1, 0x26, 0xd8, 0xb0, 0x71, 0x51, 0x39, 
                0x8d, 0xba, 0xf2, 0x67, 0x59, 0xbd, 0x68, 0xf9, 
                0xba, 0x84, 0xd1, 0xb6, 0xd8, 0xdd, 0xaf, 0x5f, 
                0x40, 0x72, 0x91, 0x3b, 0x69, 0x69, 0xe1, 0x0c, 
                0x64, 0xd5, 0x9e, 0x4d, 0xbb, 0x76, 0x47, 0xe0, 
                0x5c, 0x6f, 0xbd, 0x22, 0x98, 0x7b, 0x19, 0x99, 
                0x62, 0xcc, 0x49, 0x0f, 0xd3, 0x12, 0x8b, 0xb7, 
                0xaf, 0x81, 0xb3, 0x55, 0xfd, 0xfa, 0x98, 0xac, 
                0x6c, 0x70, 0xe6, 0xc5, 0x0b, 0x58, 0xaf, 0xde, 
                0xcc, 0xb9, 0x6b, 0x6c, 0xd9, 0x96, 0xf7, 0x0e, 
                0x25, 0x5d, 0xfa, 0x76, 0xe4, 0xd0, 0x7f, 0x47, 
                0xfb, 0xcc, 0xcb, 0x12, 0x78, 0x83, 0xc9, 0xa8, 
                0x79, 0xe7, 0x7e, 0xf0, 0xb8, 0x47, 0xe5, 0xba, 
                0xb0, 0x5b, 0x27, 0x26, 0xcc, 0x7c, 0x75, 0x17, 
                0xc4, 0xd1, 0x75, 0x03, 0xc7, 0x9d, 0xdb, 0x36, 
                0x6f, 0xc8, 0x2b, 0x89, 0x62, 0xef, 0xcd, 0xfd, 
                0x78, 0x78, 0xed, 0x56, 0xb1, 0x0f, 0xff, 0xce, 
                0xe3, 0xad, 0xc8, 0xd9, 0xce, 0x41, 0x4b, 0xf7, 
                0xdc, 0xde, 0xb5, 0xef, 0xc8, 0xe8, 0xe7, 0x0b, 
                0xb7, 0x2f, 0xbe, 0x2c, 0xd0, 0xef, 0xf7, 0xfe, 
                0xf1, 0x83, 0xf7, 0x8f, 0x5b, 0x70, 0xfc, 0x95, 
                0x37, 0x60, 0x20, 0xaa, 0xd9, 0xc5, 0xde, 0x7b, 
                0xd5, 0xd1, 0x36, 0x1d, 0x75, 0xb1, 0xfd, 0x66, 
                0x1e, 0x80, 0x04, 0xe6, 0x57, 0x20, 0x80, 0x01, 
                0x8e, 0x27, 0x60, 0x85, 0x16, 0x92, 0x36, 0x5e, 
                0x84, 0x38, 0xa8, 0x47, 0xc1, 0x82, 0xd0, 0xb9, 
                0xd7, 0x19, 0x7b, 0x24, 0x5a, 0x87, 0x96, 0x4c, 
                0x2a, 0xee, 0xb6, 0x9d, 0x5f, 0xda, 0xf1, 0xc7, 
                0x22, 0x8a, 0xff, 0xc1, 0x78, 0x9f, 0x63, 0xc2, 
                0xd9, 0x28, 0x9a, 0x0d, 0x20, 0xee, 0x04, 0x5f, 
                0x4e, 0x21, 0xc5, 0xd7, 0xe3, 0x88, 0x0d, 0x76, 
                0x28, 0xa3, 0x7f, 0x32, 0xfe, 0x57, 0xa4, 0x8b, 
                0x10, 0xc6, 0x48, 0x1e, 0x8e, 0xe2, 0x1d, 0x69, 
                0xda, 0x04, 0xc5, 0xdd, 0xb0, 0x23, 0x5d, 0xcf, 
                0x41, 0x00, 0x9f, 0x88, 0x22, 0xc2, 0xb5, 0xa0, 
                0x7e, 0x6b, 0xc1, 0x44, 0x9e, 0x91, 0x4a, 0x22, 
                0x89, 0x95, 0x93, 0x18, 0x66, 0xd8, 0x5b, 0x98, 
                0xe7, 0xc5, 0xe8, 0x43, 0x95, 0x11, 0x04, 0x89, 
                0xe0, 0x95, 0x0a, 0xca, 0x39, 0xa7, 0x66, 0x1d, 
                0xda, 0xa6, 0x61, 0x8b, 0xc6, 0x21, 0x79, 0x24, 
                0x8c, 0x51, 0xee, 0x59, 0x61, 0x9e, 0x63, 0xf6, 
                0xf7, 0xe1, 0x81, 0x0c, 0x0e, 0xb9, 0x14, 0x07, 
                0xb3, 0x6d, 0x79, 0x62, 0x9f, 0x37, 0x7a, 0x78, 
                0xe6, 0x9e, 0x0c, 0xd0, 0xb8, 0x1f, 0x8b, 0x94, 
                0xee, 0x46, 0x23, 0x9a, 0x1c, 0xea, 0xb0, 0x9c, 
                0x7c, 0x99, 0x21, 0xc6, 0x23, 0x9d, 0x87, 0xda, 
                0x09, 0x1b, 0x9c, 0x1b, 0xe6, 0x58, 0xdf, 0x84, 
                0xe0, 0xa9, 0x79, 0x2a, 0xa4, 0x4e, 0x66, 0x3a, 
                0x28, 0x9b, 0xb0, 0x16, 0xda, 0xa8, 0x65, 0x9d, 
                0x01, 0xf9, 0x26, 0xa2, 0x73, 0x3a, 0x38, 0x5b, 
                0xaa, 0x8e, 0xe6, 0xf8, 0xa7, 0x84, 0x93, 0x5e, 
                0x8a, 0xaa, 0xa5, 0xc4, 0x3e, 0xf9, 0xfe, 0x2b, 
                0xa1, 0xc9, 0x02, 0x4b, 0x43, 0xa7, 0x26, 0x8a, 
                0xea, 0xa9, 0x90, 0xbc, 0x1a, 0x26, 0x22, 0xa4, 
                0xad, 0x0e, 0x5b, 0xa3, 0x56, 0xac, 0x3a, 0x60, 
                0xa6, 0x07, 0x9b, 0x22, 0xf7, 0x6d, 0x0d, 0xce, 
                0xea, 0xfa, 0x29, 0x90, 0xa4, 0x6e, 0x46, 0x9d, 
                0x8f, 0xb6, 0x9e, 0xaa, 0x2a, 0x9b, 0x49, 0x1a, 
                0xb7, 0xa6, 0xb2, 0xdc, 0x1a, 0xeb, 0x6e, 0x06, 
                0x53, 0xbe, 0xab, 0xa3, 0xa1, 0x5c, 0x3e, 0x6b, 
                0x18, 0x8f, 0x1b, 0xe8, 0xfb, 0x6b, 0xbb, 0xcc, 
                0xea, 0x29, 0xac, 0xa6, 0xab, 0x76, 0x2b, 0xa9, 
                0x8b, 0x14, 0x6c, 0xab, 0xaa, 0x81, 0xb5, 0x32, 
                0x2a, 0x2a, 0x63, 0x11, 0x5b, 0xb9, 0xe5, 0x85, 
                0x83, 0xc6, 0x4b, 0x61, 0xbd, 0xcb, 0x7a, 0xc7, 
                0x24, 0xb1, 0x09, 0xeb, 0x99, 0x1c, 0xc1, 0xe2, 
                0xea, 0xbb, 0x2e, 0x97, 0x1e, 0x98, 0x4a, 0x71, 
                0xc5, 0x21, 0xdf, 0xc9, 0x27, 0xbe, 0x50, 0x62, 
                0x2c, 0x28, 0xc7, 0x94, 0xb2, 0x1c, 0xae, 0xc7, 
                0xb4, 0xae, 0x5b, 0x32, 0xb9, 0x16, 0xe8, 0x9c, 
                0xab, 0x7a, 0xe1, 0x8e, 0x09, 0x33, 0x91, 0xf3, 
                0xfa, 0xd9, 0xb0, 0xc0, 0x2b, 0xb7, 0xfa, 0xaa, 
                0xc5, 0x54, 0x92, 0x0c, 0x31, 0xca, 0xed, 0xbd, 
                0x37, 0xd6, 0x8f, 0xd8, 0xba, 0xfb, 0x73, 0xb0, 
                0x2f, 0x83, 0xab, 0x74, 0xa4, 0x46, 0x0f, 0x0c, 
                0x28, 0xd5, 0x82, 0x8e, 0xfc, 0x70, 0xb4, 0xa3, 
                0x4e, 0xbc, 0x6f, 0x83, 0x09, 0x96, 0x0a, 0xf0, 
                0xd5, 0x9a, 0x0e, 0x0d, 0x2c, 0xc3, 0x45, 0xb3, 
                0x8d, 0xac, 0xd0, 0x11, 0x1c, 0x0d, 0x73, 0xcb, 
                0xf9, 0x5e, 0x94, 0x48, 0x46, 0x71, 0xe0, 0x9d, 
                0x89, 0xde, 0x68, 0xf0, 0x9d, 0xb7, 0xdf, 0x67, 
                0x00, 0x4e, 0x88, 0xe0, 0x83, 0x13, 0xfe, 0x89, 
                0xe1, 0x58, 0x20, 0x9e, 0xb8, 0xe2, 0x56, 0x30, 
                0xde, 0x87, 0xe3, 0x57, 0x40, 0xad, 0x5e, 0x8a, 
                0xe4, 0x55, 0x50, 0xbe, 0x87, 0xe5, 0x97, 0x63, 
                0x5e, 0x87, 0xe6, 0x54, 0x70, 0xde, 0xb9, 0xe7, 
                0x52, 0x80, 0xbe, 0x8a, 0xe8, 0x51, 0x90, 0x3e, 
                0x87, 0xe9, 0xa7, 0xa3, 0x3e, 0x92, 0xea, 0x4f, 
                0xb0, 0xde, 0xba, 0xeb, 0x4d, 0xc0, 0x5e, 0x88, 
                0xec, 0xb3, 0xd3, 0x0e, 0x95, 0xed, 0x4a, 0xe0, 
                0x9e, 0xbb, 0xee, 0x48, 0xf0, 0x2e, 0xb5, 0xef, 
                0xbf, 0x03, 0x9f, 0xb6, 0xf0, 0x40, 0x10, 0x5f, 
                0xbc, 0xf1, 0x6d, 0x22, 0x5f, 0xb2, 0xf2, 0x7a, 
                0x30, 0xbf, 0xa5, 0xf3, 0xcb, 0x43, 0x7f, 0xb6, 
                0xf4, 0x3b, 0x50, 0x5f, 0xbd, 0xf5, 0x9c, 0x62, 
                0xef, 0xb4, 0xf6, 0xca, 0x71, 0x4f, 0xb6, 0xf7, 
                0xd7, 0x83, 0x2f, 0x3e, 0x11, 0xe4, 0x97, 0x6f, 
                0x3e, 0xf6, 0xe8, 0xa7, 0x4f, 0xfd, 0xfa, 0xec, 
                0x23, 0xef, 0x7e, 0x11, 0xcc, 0xc7, 0x2f, 0x3f, 
                0xf1, 0xf4, 0x1b, 0x01, 0xfc, 0xfd, 0xc3, 0xb3, 
                0xae, 0x7f, 0x12, 0xa4, 0xf7, 0x7f, 0x3b, 0xc6, 
                0x01, 0xf0, 0x73, 0x07, 0x19, 0xa0, 0x01, 0x0f, 
                0x88, 0xc0, 0x04, 0x2a, 0x70, 0x81, 0x0c, 0x6c, 
                0xa0, 0x03, 0x1f, 0x08, 0xc1, 0x08, 0x4a, 0x70, 
                0x82, 0x14, 0xac, 0xa0, 0x05, 0x2f, 0x88, 0x41, 
                0xcd, 0x15, 0x00, 0x00, 0x3b
        };

        static int pos;
        static int remain;
        int len;

        /* initialize state upon first call */
        if (!seq) {
                pos = 0;
                remain = sizeof(hdwireless_gif);
        }

        /* no more data? */
        if (!remain)
                return 0;

        /* send as much data as we can */
        if (remain > rsplen)
                len = rsplen;
        else
                len = remain;
        
        memcpy(rspbuf, hdwireless_gif + pos, len);
        pos += len;
        remain -= len;
        return len;
}

/**
  * @brief  Main program.
  * @param  None
  * @retval None
  */
int main(void)
{
        USART_InitTypeDef USART_InitStructure;
        GPIO_InitTypeDef GPIO_InitStructure;
        NVIC_InitTypeDef NVIC_InitStructure;
        
        struct owlsh_ctx *sh;
        int err;
        struct httpd_ctx *httpd;

        /* GCC startup code does not currently call SystemInit() */
        SystemInit();

        SysTick_Config(SystemCoreClock / 100);

        /*!< At this stage the microcontroller clock setting is already
          configured, this is done through SystemInit() function. To reconfigure
          the default setting of SystemInit() function, refer to
          system_stm32f10x.c file */     
        
        /* Initialize LEDs, Key Button, LCD and COM port(USART) available on
           STM3210X-EVAL board */
        STM_EVAL_LEDInit(LED1);
        STM_EVAL_LEDInit(LED2);
        STM_EVAL_LEDInit(LED3);
        STM_EVAL_LEDInit(LED4);

	/* If STM32F4xG-EVAL is used then USART1 port cant 
           be used because of shared MCU-pins 
	   for Micro-SD and RS232-port. 
           We map to USART3 instead.*/
#ifdef USE_STM324xG_EVAL

        /* Enable GPIO clock */
 	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOB, ENABLE);
  
    	/* Enable UART clock */
    	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART3, ENABLE);
 
  	/* Connect PB10 to USART3_Tx*/
  	GPIO_PinAFConfig(GPIOB, GPIO_PinSource10, GPIO_AF_USART3);

  	/* Connect PB11 to USART3_Rx*/
  	GPIO_PinAFConfig(GPIOB, GPIO_PinSource11, GPIO_AF_USART3);

  	/* Configure USART Tx as alternate function  */
  	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;

  	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_10;
  	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  	GPIO_Init(GPIOB, &GPIO_InitStructure);

 	 /* Configure USART Rx as alternate function  */
  	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
  	GPIO_InitStructure.GPIO_Pin = GPIO_Pin_11;
  	GPIO_Init(GPIOB, &GPIO_InitStructure);
        
        /* Configure USART RX interrupt */
        NVIC_InitStructure.NVIC_IRQChannel = USART3_IRQn;
        NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
        NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
        NVIC_Init(&NVIC_InitStructure);

        /* Configure interrupt controller */
        /* Move this to platform init file */
        /* Configure the NVIC Preemption Priority Bits */  
        NVIC_PriorityGroupConfig(NVIC_PriorityGroup_0);

        if (owluart_init(USART3,
                         UART_BAUDRATE_115200,
                         UART_DATABITS_8,
                         UART_PARITY_NONE,
                         UART_STOPBITS_1, 
                         0) < 0)
                return -1;
#else
         /* Enable GPIO clock */
        RCC_APB2PeriphClockCmd(EVAL_COM1_TX_GPIO_CLK | EVAL_COM1_CLK |
                               RCC_APB2Periph_AFIO, ENABLE);
        
        /* Configure USART Tx as alternate function push-pull */
        GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF_PP;
        GPIO_InitStructure.GPIO_Pin = EVAL_COM1_TX_PIN;
        GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
        GPIO_Init(EVAL_COM1_TX_GPIO_PORT, &GPIO_InitStructure);
        
        /* Configure USART Rx as input floating */
        GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
        GPIO_InitStructure.GPIO_Pin = EVAL_COM1_RX_PIN;
        GPIO_Init(EVAL_COM1_RX_GPIO_PORT, &GPIO_InitStructure);
  
        /* Configure USART RX interrupt */
        NVIC_InitStructure.NVIC_IRQChannel = EVAL_COM1_IRQn;
        NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
        NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
        NVIC_Init(&NVIC_InitStructure);

        /* Configure interrupt controller */
        /* Move this to platform init file */
        /* Configure the NVIC Preemption Priority Bits */  
        NVIC_PriorityGroupConfig(NVIC_PriorityGroup_0);

        if (owluart_init(EVAL_COM1, 
                         UART_BAUDRATE_115200,
                         UART_DATABITS_8,
                         UART_PARITY_NONE,
                         UART_STOPBITS_1, 
                         0) < 0)
                return -1;
#endif

        /* Initialize the LCD */
#ifdef USE_STM324xG_EVAL
	STM324xG_LCD_Init();
#elif defined (USE_STM32100B_EVAL)
        STM32100B_LCD_Init();
#elif defined (USE_STM3210B_EVAL)
        STM3210B_LCD_Init();
#elif defined (USE_STM3210E_EVAL)
        STM3210E_LCD_Init();
#elif defined (USE_STM3210C_EVAL)
        STM3210C_LCD_Init();
#elif defined (USE_STM32100E_EVAL)
        STM32100E_LCD_Init();  
#endif
      
        /* Display message on LCD */
        
        /* Clear the LCD */ 
        LCD_Clear(LCD_COLOR_WHITE);
        /* Set the LCD Back Color */
        LCD_SetBackColor(LCD_COLOR_BLUE);
        /* Set the LCD Text Color */
        LCD_SetTextColor(LCD_COLOR_WHITE);
        LCD_DisplayStringLine(LCD_LINE_0, (uint8_t *)MESSAGE1);
        LCD_DisplayStringLine(LCD_LINE_1, (uint8_t *)MESSAGE2);
        LCD_DisplayStringLine(LCD_LINE_2, (uint8_t *)MESSAGE3);

        /* Appplication code here */

#if defined(APP_MODE_WIFI)
        owltmr_init();

        /*
         * Create a shell on the COM1 where we can run our wifi
         * test commands.
         *
         */
        if ((sh = owlsh_create(NULL, owluart_write, EVAL_COM1, 0)) == NULL) {
                owl_err("could not create shell");
                return -1;
        }

        /*
         * Initialize and start the Wi-Fi device.
         *
         */
        wl_conf_alignment(16);

# if defined(OWL_CONFIG_FW_AP)
        if ((err = owlap_init(fw_read_cb, NULL, link_status_cb, sh))) {
                owl_err("ap init failed (%d)", err);
                return -1;
        }
# else
        if ((err = owlwifi_init(fw_read_cb, NULL, link_status_cb, sh))) {
                owl_err("wifi init failed (%d)", err);
                return -1;
        }
# endif


        /*
         * Initialize and start the lwip TCP/IP stack
         *
         */
        if (owlnet_init(ip_status_cb, sh) == NULL) {
                owl_err("net init failed");
                return -1;
        }

        /* Add console commands */
#if defined(OWL_CONFIG_FW_LEGACY)
        owlsh_addcmd(sh, &cmd_scan, NULL);
        owlsh_addcmd(sh, &cmd_connect, NULL);
        owlsh_addcmd(sh, &cmd_wpass, NULL);
        owlsh_addcmd(sh, &cmd_powersave, NULL);
        owlsh_addcmd(sh, &cmd_psconf, NULL);
        owlsh_addcmd(sh, &cmd_setkey, NULL);
        owlsh_addcmd(sh, &cmd_dhcp, NULL);
#elif defined(OWL_CONFIG_FW_AP)
        owlsh_addcmd(sh, &cmd_ap, NULL);
        owlsh_addcmd(sh, &cmd_ifconfig, NULL);
        owlsh_addcmd(sh, &cmd_ttcp, NULL);
#endif
        owlsh_addcmd(sh, &cmd_help, NULL);

        /* Create the HTTP server */
        if ((httpd = httpd_create(80)) == NULL) {
                owl_err("failed to create httpd");
                return -1;
        }

        /* Add documents page to HTTP server */
        if ((httpd_add_document(httpd, HTTP_METHOD_GET, "/index.html?p=LED1",
                                httpd_led_cb, (void *) LED1)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(httpd, HTTP_METHOD_GET, "/index.html?p=LED2",
                                httpd_led_cb, (void *) LED2)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(httpd, HTTP_METHOD_GET, "/index.html?p=LED3",
                                httpd_led_cb, (void *) LED3)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(httpd, HTTP_METHOD_GET, "/index.html?p=LED4",
                                httpd_led_cb, (void *) LED4)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(httpd, HTTP_METHOD_GET, "/index.html",
                                httpd_index_cb, NULL)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(httpd, HTTP_METHOD_GET, "/image.gif",
                                httpd_logo_cb, NULL)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }

        /* Start HTTP server */
        httpd_start(httpd);
        
        /* main loop */
        for (;;) {
                if (owluart_poll(EVAL_COM1)) {
                        owlsh_key(sh, owluart_inb(EVAL_COM1, 0));
                }
                    
                /* Execute any pending timers. Note that we use timers for both
                 * the wifi and the ip stack to make forward progress, rather
                 * than calling each poll-function here.
                 */
                owltmr_poll();
        }

#else
        (void) err;
        (void) sh;
        
        /* perform wifi loopback test of 32768 packets, starting with size 32
         * incrementing by 1
         */
        wl_conf_alignment(16);
        owlloop_run(fw_read_cb, NULL, 32, 1, 32768);
        for (;;);

#endif

}
