/* 
 * Implementation of wl_api low level transport driver, as defined by wl_sdio.h
 *
 * Note that owl_sdio_mdelay() is not defined here, so it must be implemented
 * for the specific board.
 *
 */


/*
 * Note that this driver currently has these limitations:
 *
 * - Only size alignment 4 works
 * - Min transfer size seems to be > 8.
 *
 * A problem could occur due to lwip's chaining of pbuf's. We currently pad
 * the complete packet in wl_rx() to a multple of 16 (as set in
 * wl_conf_alignment()). This avoids e.g. packets of 516 bytes,
 * where the last would have been 4 bytes (too small). However, we will break
 * if any chain in the pbuf is not size aligned to 4 bytes since we cannot
 * pad individual pbufs. We will also break if a single pbuf is too small.
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <stdint.h>
#include <stdlib.h> /* for NULL */

#include "wl_sdio.h"
#include "owl/core/owl_debug.h"

#if defined(STM32F40XX)
#include "stm324xg_eval.h"
#include "stm32f4xx_sdio.h"
#include "stm32f4xx_dma.h"
#else
#include "stm32_eval.h"
#include "stm32f10x_sdio.h"
#endif

#define OWL_SDIO_DMA  /* enable DMA for tx and rx */
#define OWL_SDIO_IRQ  /* use sdio irq to detect pending rx data */
#define OWL_SDIO_4BIT /* enable 4 bit sdio */
//#define OWL_SDIO_TRACE_INIT /* enable dbg tracing during card init and fw dl */



#ifdef OWL_SDIO_TRACE_INIT
# define owl_dbg_noinit(args...) owl_dbg(args)
# define owl_hexdump_noinit(args...) owl_hexdump(args)
#else
# define owl_dbg_noinit(args...) do {                                   \
                if (initialized) {                                      \
                        owl_dbg(args);                                  \
                }                                                       \
        } while(0)
# define  owl_hexdump_noinit(args...) do {                              \
                if (initialized) {                                      \
                        owl_hexdump(args);                              \
                }                                                       \
        } while(0)
#endif 

void SDIO_IRQHandler(void);
static int initialized = 0; /* set to 1 when firmware is downloaded */


/**
 * This function will be invoked when wlan initialization should be performed,
 * this happens when the wl_fw_download() function in the transport group of
 * wl_api is invoked.
 *
 * The wifi device supports sdio high speed mode and clock frequencies up to
 * 50 MHz.
 *
 * The function is responsible for doing any necessary sdio initialization such
 * as allocating gpio's, setting up the mci master, one time allocations of
 * dma buffers etc.
 *
 * @param flags is an out parameter that should hold any sdio flags upon return.
 *        The avaible flags are prefixed with SDIO_FLAG_
 * 
 *
 */
void owl_sdio_init(uint8_t *flags)
{
        SDIO_InitTypeDef SDIO_InitStructure;

        owl_dbg("flags:%p", flags);

        SD_LowLevel_Init();
        
         /* The SDIO-clock is 48MHz. The resulting speed 
           is 48/(dividervalue+2). The dividervalue for STM32F4 
           when using SD to Micro-SD adapter is 2 resulting 
	   in 12MHz operation. The minimum dividervalue for 
           STM32F1 when using SD to Micro-SD adapter
           is 4 resulting in 8MHz operation. When not using DMA with 
           STM32F1 the dividervalue is 2.
	   Not using DMA with STM32F4 is not supported at this time. */

#ifdef OWL_SDIO_DMA

#if defined (STM32F40XX)
	SDIO_InitStructure.SDIO_ClockDiv = 2;
#else
	SDIO_InitStructure.SDIO_ClockDiv = 4;
#endif	

#else 				
	SDIO_InitStructure.SDIO_ClockDiv = 2;
#endif 
        
        SDIO_InitStructure.SDIO_ClockEdge = SDIO_ClockEdge_Rising;
        SDIO_InitStructure.SDIO_ClockBypass = SDIO_ClockBypass_Disable;

        /* NOTE: SDIO IRQ's are not properly detected if SDIO_ClockPowerSave is
           enabled. Seems like the DAT1 is only checked for pending irq when
           the clock is running */
        SDIO_InitStructure.SDIO_ClockPowerSave = SDIO_ClockPowerSave_Disable;
#ifdef OWL_SDIO_4BIT
        SDIO_InitStructure.SDIO_BusWide = SDIO_BusWide_4b;
#else
        SDIO_InitStructure.SDIO_BusWide = SDIO_BusWide_1b;
#endif
        SDIO_InitStructure.SDIO_HardwareFlowControl =
                SDIO_HardwareFlowControl_Enable;

        owl_info("clk_div:%d", SDIO_InitStructure.SDIO_ClockDiv);

        SDIO_Init(&SDIO_InitStructure);
        
        /* Set Power State to ON */
        SDIO_SetPowerState(SDIO_PowerState_ON);

        /* Enable SDIO Clock */
        SDIO_ClockCmd(ENABLE);
        
        /* Configure SDIO interrupts */
        SDIO_SetSDIOOperation(ENABLE);
        NVIC_EnableIRQ(SDIO_IRQn);

        /* Wait for SPB104 reset circuit to power the device after cold-start.
         * This is not required if HDG104 is used directly with the
         * GPIO_SHUTDOWN rather than a reset circuit
         */
        owl_sdio_mdelay(500);
        
#ifndef OWL_SDIO_IRQ
        *flags = SDIO_FLAG_POLL;
#else
        *flags = 0;
#endif
        
#ifndef OWL_SDIO_4BIT
        *flags |= SDIO_FLAG_1BIT_MODE;
#endif

        owl_dbg("*flags:0x%x", *flags);
}



/**
 * This function will be invoked when an sdio cmd should be sent to the
 * device.
 *
 * @param idx is the sdio command number
 * @param arg is the sdio command argument
 * @param flags specifies other options, such as any transfer direction.
 * @param rsp should hold the command response upon return. If null, the
 *        response can be ignored.
 * @param data holds a pointer to any data that might follow the command. This
 *        allows the sdio driver to setup dma transfers while waiting for the
 *        command response. NULL if no data transfer will follow. Note that
 *        the same data pointer will be passed to owl_sdio_tx(), which should
 *        start the actual transfer.
 * @param len is the length of the data buffer.
 *
 */

/* SD_LowLevel_DMA_RxConfig does not allow const data, so disable cast-qual
   here */
#pragma GCC diagnostic ignored "-Wcast-qual" 
void owl_sdio_cmd(uint8_t idx, uint32_t arg, uint8_t flags, uint32_t *rsp,
                    const uint8_t *data, uint16_t len)
{
        uint32_t status;
        uint32_t respcmd; /* SDIO->RESPCMD */
        uint32_t resp1;   /* SDIO->RESP1   */

        SDIO_CmdInitTypeDef SDIO_CmdInitStructure;
        SDIO_DataInitTypeDef SDIO_DataInitStructure;   

        owl_dbg_noinit("idx:0x%x arg:0x%lx flags:%x", idx, arg, flags);
        owl_dbg_noinit("rsp:%p data:%p len:%d", rsp, data, len);

        /* setup any data transfer */
        if (data != NULL) {
                SDIO_DataInitStructure.SDIO_DataTimeOut = 0xFFFFF;
                SDIO_DataInitStructure.SDIO_DataLength = len;
                SDIO_DataInitStructure.SDIO_DataBlockSize =
                        SDIO_DataBlockSize_1b;
                SDIO_DataInitStructure.SDIO_TransferMode =
                        SDIO_TransferMode_Stream;
                SDIO_DataInitStructure.SDIO_DPSM = SDIO_DPSM_Enable;
                
                if (flags & CMD_FLAG_TO_HOST)
                        SDIO_DataInitStructure.SDIO_TransferDir =
                                SDIO_TransferDir_ToSDIO;
                else
                        SDIO_DataInitStructure.SDIO_TransferDir =
                                SDIO_TransferDir_ToCard;

                SDIO_DataConfig(&SDIO_DataInitStructure);

#ifdef OWL_SDIO_DMA
                if (flags & CMD_FLAG_TO_HOST) {

                	/* Prepare RX-DMA here since if we take too long after
                           sending the command but before enabling DMA we might
                           fill up the fifo (128b) and lose data. This would
                           usually happen with traces enabled */
				
#if defined(STM32F40XX) /* STM32F40XX DMA RX */
                        uint32_t * MemoryDST = (uint32_t*)data;

			DMA_InitTypeDef SDDMA_InitStructure;			

  			DMA_ClearFlag(SD_SDIO_DMA_STREAM, 
                                      SD_SDIO_DMA_FLAG_FEIF | 
                                      SD_SDIO_DMA_FLAG_DMEIF | 
                                      SD_SDIO_DMA_FLAG_TEIF | 	
                                      SD_SDIO_DMA_FLAG_HTIF | 
                                      SD_SDIO_DMA_FLAG_TCIF);

  			/* DMA2 Stream3 disable */
  			DMA_Cmd(SD_SDIO_DMA_STREAM, DISABLE);

  			/* DMA2 Stream3 Deinit */
  			DMA_DeInit(SD_SDIO_DMA_STREAM);
	
			/* Setup DMA control registers.
		           Peripheral base adress is SDIO fifo, 
                           Direction is from SDIO to Memory,
		           Memory pointer is incremented, 
                           fifo size is 32 bits (one word),
		   	   each burst is 32 bits (one word). 
                           BufferSize is not used when using 
                           SDIO flowcontrol. */

  			SDDMA_InitStructure.DMA_Channel = SD_SDIO_DMA_CHANNEL;
  			SDDMA_InitStructure.DMA_PeripheralBaseAddr = 
                                (uint32_t)SDIO_FIFO_ADDRESS;
  			SDDMA_InitStructure.DMA_Memory0BaseAddr = 
                                (uint32_t) MemoryDST;
  			SDDMA_InitStructure.DMA_DIR = 
                                DMA_DIR_PeripheralToMemory;
  			SDDMA_InitStructure.DMA_BufferSize = 
                                len / 4;	 
  			SDDMA_InitStructure.DMA_PeripheralInc = 
                                DMA_PeripheralInc_Disable;
  			SDDMA_InitStructure.DMA_MemoryInc = 
                                DMA_MemoryInc_Enable;
  			SDDMA_InitStructure.DMA_PeripheralDataSize = 
                                DMA_PeripheralDataSize_Word;
  			SDDMA_InitStructure.DMA_MemoryDataSize = 
                                DMA_MemoryDataSize_Word;
  			SDDMA_InitStructure.DMA_Mode = 
                                DMA_Mode_Normal;
  			SDDMA_InitStructure.DMA_Priority = 
                                DMA_Priority_High;
  			SDDMA_InitStructure.DMA_FIFOMode = 
                                DMA_FIFOMode_Enable;
  			SDDMA_InitStructure.DMA_FIFOThreshold = 
                                DMA_FIFOThreshold_Full;
  			SDDMA_InitStructure.DMA_MemoryBurst = 
                                DMA_MemoryBurst_INC4;
  			SDDMA_InitStructure.DMA_PeripheralBurst = 
                                DMA_PeripheralBurst_INC4;  			
			
			DMA_Init(SD_SDIO_DMA_STREAM, &SDDMA_InitStructure);
  
  			DMA_FlowControllerConfig(SD_SDIO_DMA_STREAM, 
                                                 DMA_FlowCtrl_Peripheral);    

  			/* DMA2 Stream3 enable */
  			DMA_Cmd(SD_SDIO_DMA_STREAM, ENABLE);
			
			/* Start DMA transfer */
		        SDIO_DMACmd(ENABLE);
			
#else /* STM32F1 DMA RX */
			SDIO_DMACmd(ENABLE);
                        SD_LowLevel_DMA_RxConfig((uint32_t *) data, len);
#endif		
                }
#endif
        }

        /* send the command */
        SDIO_CmdInitStructure.SDIO_Argument = arg;
        SDIO_CmdInitStructure.SDIO_CmdIndex = idx;
        SDIO_CmdInitStructure.SDIO_Response = SDIO_Response_Short;
        SDIO_CmdInitStructure.SDIO_Wait = SDIO_Wait_No;
        SDIO_CmdInitStructure.SDIO_CPSM = SDIO_CPSM_Enable;
        SDIO_SendCommand(&SDIO_CmdInitStructure);
        
        /* wait until response is obtained or an error occurs */
        status = SDIO->STA;
        while (!(status & (SDIO_FLAG_CCRCFAIL | SDIO_FLAG_CTIMEOUT |
                           SDIO_FLAG_CMDREND))) {
                status = SDIO->STA;
        }

        /* check for transfer errors */
        
        if (status & SDIO_FLAG_CTIMEOUT) {
                SDIO_ClearFlag(SDIO_FLAG_CTIMEOUT);
                owl_err("ctimeout");
                owl_err("idx:0x%x arg:0x%lx flags:0x%x", idx, arg, flags);
                owl_err("rsp:%p data:%p len:%d", rsp, data, len);
                return;
                
        } else if (status & SDIO_FLAG_CCRCFAIL) {
                SDIO_ClearFlag(SDIO_FLAG_CCRCFAIL);
                owl_err("ccrcfail");
                owl_err("idx:0x%x arg:0x%lx flags:0x%x", idx, arg, flags);
                owl_err("rsp:%p data:%p len:%d", rsp, data, len);
                return;
        }

        /* Get the command index from the response.
         *
         * NOTE: we will get response timeout for cmd52 if we don't read
         * this register first!
         */
        respcmd = SDIO_GetCommandResponse();
        if (respcmd != idx) {
                owl_err("got response from wrong command:%lu expected:%d",
                        respcmd, idx);
                return;
        }

        /* read the response */
        resp1 = SDIO_GetResponse(SDIO_RESP1);
        SDIO_ClearFlag(SDIO_FLAG_CMDREND);

        owl_dbg_noinit("respcmd:0x%02lx resp1:0x%08lx", respcmd, resp1);
        
        if (rsp != NULL)
                *rsp = resp1;
}
#pragma GCC diagnostic warning "-Wcast-qual"


/**
 * This function will be invoked when data should be transmitted to the device.
 *
 * If wl_conf_alignment() was called with the size_align parameter set to
 * non-zero, the pad parameter should be used. If the pad parameter is not 0,
 * additional data must be transmitted after the data buffer has be sent.
 * Depending on how the data buffer was first allocated (probably by an TCP/IP
 * stack), it might be safe or unsafe to continue reading beyond the data
 * buffer to transmit the additional padding bytes.
 *
 * @param data holds a pointer to the data to transmit, the pointer is the
 *            same as the one passed to wl_tx().
 * @param len is the number of bytes that should be transmitted, including
 *            padding.
 * @param pad is the number of padding bytes to send.
 *
 */

/* SD_LowLevel_DMA_TxConfig does not allow const data, so disable cast-qual
   here */
#pragma GCC diagnostic ignored "-Wcast-qual" 
void owl_sdio_tx(const uint8_t *data, uint16_t len, uint8_t pad)
{
        int i;
        uint8_t remain; /* bytes remaining in data array, not including pad */
        uint8_t tail[16]; /* put remaining bytes and the pad in this buffer */
        uint16_t datalen = len - pad; /* size of data array, not includ pad */

        /* the asserts below will hold if wl_conf_alignment() was set to 16 */
        owl_assert(len % 4 == 0); /* SDIO host can not handle non-size aligned
                                     data */
        owl_assert(len > 8); /* SDIO host can not handle buffers that are 8
                                bytes or smaller */
        owl_assert(pad < 16);
        owl_dbg_noinit("data:%p len:%d pad:%d", data, len, pad);

#ifndef OWL_SDIO_DMA
        const uint32_t *data32 = (const uint32_t *) data;
        
        for (i = 0; i < datalen / 4;) {
                uint32_t status;
                status = SDIO->STA;
                if (status & (SDIO_FLAG_DTIMEOUT | SDIO_FLAG_DCRCFAIL |
                              SDIO_FLAG_TXUNDERR | SDIO_FLAG_STBITERR |
                              SDIO_FLAG_RXOVERR)) {
                        owl_err("status:0x%08lx", status);
                        return;
                }

                /* don't write if fifo is full */
                if (status & SDIO_FLAG_TXFIFOF)
                        continue;
                                
                /* write 4 bytes at a time */
                SDIO_WriteData(data32[i++]);
        }

#else /* OWL_SDIO_DMA */
        if (datalen / 4 > 0) {
               	
#if defined(STM32F40XX)
		uint32_t * MemorySRC = (uint32_t*)data;		
		uint32_t BufferSize = 4 * (datalen / 4);
	     
 		DMA_InitTypeDef SDDMA_InitStructure;

  		DMA_ClearFlag(SD_SDIO_DMA_STREAM, 
                              SD_SDIO_DMA_FLAG_FEIF | 
                              SD_SDIO_DMA_FLAG_DMEIF | 
                              SD_SDIO_DMA_FLAG_TEIF | 
                              SD_SDIO_DMA_FLAG_HTIF | 
                              SD_SDIO_DMA_FLAG_TCIF);
		
		/* DMA2 Stream3 disable */
  		DMA_Cmd(SD_SDIO_DMA_STREAM, DISABLE);
		
		/* DMA2 Stream3 Deinit */
  		DMA_DeInit(SD_SDIO_DMA_STREAM);
	
		/* Setup DMA control registers.
		   Peripheral base adress is SDIO fifo, 
                   Direction is from memory to SDIO,
		   Memory pointer is incremented, 
                   fifo size is 32 bits (one word),
		   each burst is 32 bits (one word). 
                   BufferSize is not used when using SDIO flowcontrol. */

  		SDDMA_InitStructure.DMA_Channel = 
                        SD_SDIO_DMA_CHANNEL;
  		SDDMA_InitStructure.DMA_PeripheralBaseAddr = 
                        (uint32_t)SDIO_FIFO_ADDRESS;   
  		SDDMA_InitStructure.DMA_Memory0BaseAddr = 
                        (uint32_t)MemorySRC ;  	    
  		SDDMA_InitStructure.DMA_DIR = 
                        DMA_DIR_MemoryToPeripheral;	 	    
  		SDDMA_InitStructure.DMA_BufferSize = 
                        BufferSize / 4;          	            
  		SDDMA_InitStructure.DMA_PeripheralInc = 
                        DMA_PeripheralInc_Disable;         
  		SDDMA_InitStructure.DMA_MemoryInc = 
                        DMA_MemoryInc_Enable;   		   
  		SDDMA_InitStructure.DMA_PeripheralDataSize = 
                        DMA_PeripheralDataSize_Word;   
  		SDDMA_InitStructure.DMA_MemoryDataSize = 
                        DMA_MemoryDataSize_Word;	    
  		SDDMA_InitStructure.DMA_Mode = 
                        DMA_Mode_Normal;
  		SDDMA_InitStructure.DMA_Priority = 
                        DMA_Priority_High;
  		SDDMA_InitStructure.DMA_FIFOMode = 
                        DMA_FIFOMode_Enable; 
  	        SDDMA_InitStructure.DMA_FIFOThreshold = 
                        DMA_FIFOThreshold_Full;
  		SDDMA_InitStructure.DMA_MemoryBurst = 
                        DMA_MemoryBurst_INC4;
  		SDDMA_InitStructure.DMA_PeripheralBurst = 
                        DMA_PeripheralBurst_INC4;
  		
                DMA_Init(SD_SDIO_DMA_STREAM, &SDDMA_InitStructure);
                 
  		DMA_FlowControllerConfig(SD_SDIO_DMA_STREAM, 
                                         DMA_FlowCtrl_Peripheral);		
  		
                DMA_Cmd(SD_SDIO_DMA_STREAM, ENABLE);
		
		/* Start DMA transfer */
		SDIO_DMACmd(ENABLE);
		
		/* Check Stream 3 transfer complete flag */
	        while ((uint32_t)DMA_GetFlagStatus(DMA2_Stream3,
                                                   DMA_FLAG_TCIF3)== RESET);
		
#else /* STM32F1 DMA TX */	
		SD_LowLevel_DMA_TxConfig((uint32_t *) data, 4 * (datalen / 4));
                SDIO_DMACmd(ENABLE);
                while ((SD_DMAEndOfTransferStatus() == RESET));
#endif	
        }
 
#endif /* OWL_SDIO_DMA */

        /* transfer remaining bytes and padding */
        remain = datalen % 4;
        if (remain + pad > 0) {
                int offset = datalen - remain;
                uint32_t *tail32 = (uint32_t *) tail;

                owl_info("remain:%d offset:%d", remain, offset);

                for (i = 0; i < remain; i++)
                        tail[i] = data[offset + i];

                owl_assert((remain + pad) <= sizeof(tail));
                owl_assert((remain + pad) % 4 == 0);
                
                for (i = 0; i < (remain + pad) / 4; ) {
                        uint32_t status;
                        status = SDIO->STA;
                        if (status & (SDIO_FLAG_DTIMEOUT | SDIO_FLAG_DCRCFAIL |
                                      SDIO_FLAG_TXUNDERR | SDIO_FLAG_STBITERR |
                                      SDIO_FLAG_RXOVERR)) {
                                owl_err("status:0x%08lx", status);
                                return;
                        }

                        /* don't write if fifo is full */
                        if (status & SDIO_FLAG_TXFIFOF)
                                continue;

                        /* write 4 bytes at a time */
                        SDIO_WriteData(tail32[i++]);
                }
        }
        
        /* wait until transfer is done */
        owl_dbg_noinit("wait for txact");
        while (SDIO_GetFlagStatus(SDIO_FLAG_TXACT) != RESET);

        /* transfer is done */
        owl_assert(SDIO_GetFlagStatus(SDIO_FLAG_DATAEND) == SET);

        SDIO_ClearFlag(SDIO_FLAG_DATAEND);

        /* only hexdump if we're done downloading firmware */
        owl_hexdump_noinit("SDIOTX", (const char *) data, len);
}
#pragma GCC diagnostic warning "-Wcast-qual"


/**
 * This function will be invoked when data should be received from the device.
 *
 * @param data should hold the read data upon return.
 * @param len is the number of bytes to read.
 *
 */
void owl_sdio_rx(uint8_t *data, uint16_t len)
{
        owl_dbg("data:%p len:%d", data, len);
        owl_assert(len % 4 == 0);

#ifndef OWL_SDIO_DMA
        int i = 0;
        uint32_t *data32 = (uint32_t *) data;

        while (i < len / 4) {

                uint32_t status;
                status = SDIO->STA;
                if (status & (SDIO_FLAG_DTIMEOUT | SDIO_FLAG_DCRCFAIL |
                              SDIO_FLAG_TXUNDERR | SDIO_FLAG_STBITERR |
                              SDIO_FLAG_RXOVERR)) {
                        owl_err("status:0x%08lx", status);
                        return;
                }

                /* check if there is at least one word of data in the fifo */
                if (!(status & SDIO_FLAG_RXDAVL)) {
                        owl_info("rxdavl not set");
                        continue;
                }

                /* read one word of data */
                data32[i++] = SDIO_ReadData();
        }

#else /* OWL_SDIO_DMA */
        (void) len;
        (void) data; /* DMA was started in owl_sdio_cmd() */
        
#if defined(STM32F40XX) /* STM32F4 DMA RX */
	/* Check Stream 3 transfer complete flag */        
	while ((uint32_t)DMA_GetFlagStatus(DMA2_Stream3,
                                           DMA_FLAG_TCIF3) == RESET);
#else
	while ((SD_DMAEndOfTransferStatus() == RESET));
#endif
        
#endif /* OWL_SDIO_DMA */

#ifdef OWL_SDIO_4BIT
        /* NOTE: In SDIO 4-bit mode, the CPU seems to interpret received data
           on DAT1 as an interrupt and will set the SDIOIT bit when
           receiving data (not when sending). Since owl_sdio_irq() is
           called with enable=1 when reception is done, we must
           explicitly clear the pending here. If there is a _real_
           interrupt from the device just after reading is completed,
           the SDIOIT bit will be set again right after we clear it here
           since the irq line stays low as long as we don't ACK the IRQ
           on the device - so we will not lose any irq's */
        SDIO_ClearITPendingBit(SDIO_IT_SDIOIT);
#endif /* OWL_SDIO_4BIT */

        /* all data has been read so there should be no ongoing rx */
        owl_assert(SDIO_GetFlagStatus(SDIO_FLAG_RXACT) == RESET);
        owl_assert(SDIO_GetFlagStatus(SDIO_FLAG_DATAEND) == SET);

        SDIO_ClearFlag(SDIO_FLAG_DATAEND);

        owl_hexdump("SDIORX", (char *) data, len);
}



/**
 * Invoked when sdio rx interrupts from the device should be enabled or
 * disabled. 
 *
 * If SDIO_FLAG_POLL was set in wl_spi_init(), then this function can be
 * left empty.
 * 
 * @param enable specifies if interrupts should be enabled or disabled.
 *
 */
void owl_sdio_irq(uint8_t enable)
{
        owl_dbg("enable:%d SDIO->STA:0x%lx", enable, SDIO->STA);

        /* irq's are first enabled when device is initialized and firmware is
         * downloaded.
         */
        initialized = 1; 

#ifdef OWL_SDIO_IRQ

        if (enable)
                SDIO_ITConfig(SDIO_IT_SDIOIT, ENABLE);
        else
                SDIO_ITConfig(SDIO_IT_SDIOIT, DISABLE);
        
#else /* OWL_SDIO_IRQ */
        enable = enable; 
#endif /* OWL_SDIO_IRQ */
}


void SDIO_IRQHandler(void)
{
        owl_dbg("got irq SDIO->STA:0x%lx", SDIO->STA);
        SDIO_ClearITPendingBit(SDIO_IT_SDIOIT);
        wl_sdio_irq();
}
