/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_MINIMAL
#define OWL_FILE "owl_spi"

#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_list.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>


/* STM32 spi does not have support for multiple chip selects per bus, so
   we will reconfigure the bus when the chip select is changed */

struct spi_bus {
        int mode; /* master/slave */
        struct list_head list;
        SPI_TypeDef *addr;
        struct spi_device *last_device;
};

struct spi_device {
        struct spi_bus *bus;
        int cs;
        unsigned int freq;
        int xmode;
        int width;
        struct list_head list;
};

static struct list_head *devices = NULL;
static struct list_head *buses = NULL;

static void select_device(struct spi_bus *bus, struct spi_device *device);
static struct spi_bus *find_bus(SPI_TypeDef *addr);
static struct spi_device *find_device(int cs);


int
owlspi_init(void *addr, int mode)
{
        struct spi_bus *bus;

        owl_info("initializing spi bus at addr:%p to mode:%d", addr, mode);

        if (mode != OWL_SPI_MODE_MASTER) {
                owl_err("mode:%d not supported", mode);
                return OWL_ERR_NOTSUPP;
        }

        /* initialize bus list on first call */
        if (buses == NULL) {
                if ((buses = owl_os_alloc(sizeof(*buses))) == NULL) {
                        owl_err("could not alloc list");
                        return OWL_ERR_MEM;
                }

                list_init(buses);
        }

        if ((bus = find_bus(addr)) != NULL) {
                if (bus->mode == mode) {
                        owl_info("bus at addr:%p already initialized", addr);
                        return 0;

                } else {
                        owl_err("bus at addr:%p cannot change mode", addr);
                        return OWL_ERR_STATE;
                }
        }

        /* alloc bus context */
        if ((bus = owl_os_zalloc(sizeof(struct spi_bus))) == NULL) {
                owl_err("could not alloc bus");
                return OWL_ERR_MEM;
        }
        bus->addr = addr;
        bus->mode = mode;
        list_add_tail(&bus->list, buses);
        return 0;
}


int
owlspi_setup_cs(void *addr, int cs, int freq, int xmode, int width)
{
        struct spi_bus *bus;
        struct spi_device *device;

        owl_unused(cs);
        owl_info("setup cs_port:%x cs_pin:%x on addr:%p with freq:%d xmode:%d "
                 "width:%d",
                 STM32_PORT(cs), STM32_PIN(cs), addr, freq, xmode, width);

        switch (width) {
        case 8:
                break; /* ok */
        case 16:
                owl_err("width:%d not supported", width);
                return OWL_ERR_NOTSUPP;
        default:
                owl_err("invalid width:%d", width);
                return OWL_ERR_PARAM;
        };


        switch (xmode) {
        case OWL_SPI_TRANSFER_MODE_0:
        case OWL_SPI_TRANSFER_MODE_1:
        case OWL_SPI_TRANSFER_MODE_2:
        case OWL_SPI_TRANSFER_MODE_3:
                break;

        default:
                owl_err("invalid xmode:%d", xmode);
                return OWL_ERR_PARAM;
        };

        /* CS */
        owlgpio_set(cs);

        /* Make sure that the bus is initialized */
        if ((bus = find_bus((SPI_TypeDef *) addr)) == NULL) {
                owl_err("bus at addr:%p not initialized", addr);
                return OWL_ERR_STATE;
        }

        /* initialize device list on first call */
        if (devices == NULL) {
                if ((devices = owl_os_alloc(sizeof(*devices))) == NULL) {
                        owl_err("could not alloc list");
                        return OWL_ERR_MEM;
                }

                list_init(devices);
        }

        /* alloc device context */
        if ((device = owl_os_zalloc(sizeof(struct spi_device))) == NULL) {
                owl_err("could not alloc bus");
                return OWL_ERR_MEM;
        }
        device->bus = bus;
        device->width = width;
        device->xmode = xmode;
        device->freq = freq;
        device->cs = cs;
        list_add_tail(&device->list, devices);
        return 0;
}


void
owlspi_claim(void *addr, int cs)
{
        struct spi_bus *bus;
        struct spi_device *device;

        owl_unused(bus);
        owl_vdbg("addr:%p cs:%d", addr, cs);

        bus = find_bus(addr);
        device = find_device(cs);

        owl_assert(bus);
        owl_assert(device);

        select_device(bus, device);

        /* Enable SPI */
        SPI_Cmd(addr, ENABLE);

        owlgpio_clr(cs);
}

void
owlspi_release(void *addr, int cs)
{
        owl_unused(addr);
        owl_vdbg("addr:%p cs:%d", addr, cs);

        owlgpio_set(cs);

        /* Disable SPI  */
        SPI_Cmd(addr, DISABLE);
}

int
owlspi_exchange(void *bus, const void *tx, void *rx, int len)
{
        SPI_TypeDef *spi = bus;
        const uint8_t *tx8 = tx;
        uint8_t *rx8 = rx;
        int pos = 0;

        owl_assert(len);
        owl_vdbg("bus:%p tx:%p rx:%p len:%d", bus, tx, rx, len);

        if (tx) {
                owl_hexdump("SPI TX", tx, len);
        }

        /* clock out/in the bytes */
        while (pos < len) {

                while (SPI_I2S_GetFlagStatus(spi, SPI_I2S_FLAG_TXE) == RESET);

                if (tx)
                        SPI_I2S_SendData(bus, tx8[pos] & 0xff);
                else
                        SPI_I2S_SendData(bus, 0xff);

                while (SPI_I2S_GetFlagStatus(spi, SPI_I2S_FLAG_RXNE) == RESET);

                if (rx)
                        rx8[pos] = SPI_I2S_ReceiveData(spi);
                else
                        (void) SPI_I2S_ReceiveData(spi);

                pos++;
        }

        if (rx) {
                owl_hexdump("SPI RX", rx, len);
        }

        return len;
}

int
owlspi_write(void *bus, const void *buf, int len)
{
        return owlspi_exchange(bus, buf, NULL, len);
}

int
owlspi_read(void *bus, void *buf, int len)
{
        return owlspi_exchange(bus, NULL, buf, len);
}

static void
select_device(struct spi_bus *bus, struct spi_device *device)
{
        SPI_InitTypeDef SPI_InitStructure;
        unsigned int br, div;

        owl_vdbg("bus:%p device:%p", bus, device);

        /* already configured for this device? */
        if (bus->last_device == device)
                return;

        owl_dbg("bus:%p with mode:%d freq:%d xmode:%d width:%d",
                bus, bus->mode, device->freq, device->xmode, device->width);

        /* Always called when SPI is DISABLED */
        
        /* Only width 8 is supported, this is checked in owlspi_setup_cs() */
        SPI_InitStructure.SPI_DataSize = SPI_DataSize_8b;
        switch (device->xmode) {
        case OWL_SPI_TRANSFER_MODE_0:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_Low;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_1Edge;
                break;

        case OWL_SPI_TRANSFER_MODE_1:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_Low;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_2Edge;
                break;

        case OWL_SPI_TRANSFER_MODE_2:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_High;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_1Edge;
                break;

        case OWL_SPI_TRANSFER_MODE_3:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_High;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_2Edge;
                break;

        default:
                owl_assert(0);
                owl_err("invalid xmode:%d", device->xmode);
                break;
        };

        /* SPI configuration */
        SPI_InitStructure.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
        SPI_InitStructure.SPI_Mode = SPI_Mode_Master;
        SPI_InitStructure.SPI_NSS = SPI_NSS_Soft;
        SPI_InitStructure.SPI_FirstBit = SPI_FirstBit_MSB;
        SPI_InitStructure.SPI_CRCPolynomial = 7;

        for (br = 0, div = 2; br < 8; br++, div <<= 1) {
                if (SystemCoreClock / div < device->freq) {
                        owl_dbg("prescaler:%d gives div:%d and baudrate:%d",
                                br, div, SystemCoreClock / div);
                        break;
                }
        }
        SPI_InitStructure.SPI_BaudRatePrescaler = br << 3;
        SPI_Init(bus->addr, &SPI_InitStructure);

        bus->last_device = device;

        /* SPI will be set to ENABLE by owlspi_claim() */
}


static struct spi_bus *
find_bus(SPI_TypeDef *addr)
{
        struct list_head *pos = NULL;
        struct spi_bus *bus = NULL;

        list_for_each(pos, buses) {
                bus = list_entry(pos, struct spi_bus, list);
                if (bus->addr == addr)
                        return bus;
        }

        return NULL;
}

static struct spi_device *
find_device(int cs)
{
        struct list_head *pos = NULL;
        struct spi_device *device = NULL;

        list_for_each(pos, devices) {
                device = list_entry(pos, struct spi_device, list);
                if (device->cs == cs)
                        return device;
        }

        return NULL;
}
