/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_INFO

#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_uart.h>

int
owluart_init(void *uart,
             uint32_t baudrate, enum uart_databits databits,
             enum uart_parity parity, enum uart_stopbits stopbits,
             int rtscts)
{
        USART_TypeDef *port = uart;
        USART_InitTypeDef USART_InitStructure;

        owl_dbg("baudrate:%d databits:%d parity:%d stopbits:%d rtscts:%d"
                baudrate, databits, parity, stopbits, rtscts);

        if (databits != UART_DATABITS_8) {
                owl_err("databits:%d not supprted", databits);
                return OWL_ERR_NOTSUPP;
        }

        if (parity != UART_PARITY_NONE) {
                owl_err("parity:%d not supprted", parity);
                return OWL_ERR_NOTSUPP;
        }

        if (stopbits != UART_STOPBITS_1) {
                owl_err("stopbits:%d not supprted", stopbits);
                return OWL_ERR_NOTSUPP;
        }

        if (rtscts) {
                owl_err("rtscts not supprted", rtscts);
                return OWL_ERR_NOTSUPP;
        }
        
        USART_InitStructure.USART_BaudRate = baudrate;
        USART_InitStructure.USART_WordLength = USART_WordLength_8b;
        USART_InitStructure.USART_StopBits = USART_StopBits_1;
        USART_InitStructure.USART_Parity = USART_Parity_No;
        USART_InitStructure.USART_HardwareFlowControl =
                USART_HardwareFlowControl_None;
        USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
        
        /* USART configuration */
        USART_Init(port, &USART_InitStructure);
    
        /* Enable USART */
        USART_Cmd(port, ENABLE);
        return 0;
}

void
owluart_outb(void *uart, int c)
{
        USART_TypeDef *port = uart;
        
        /* write a character to the USART */
        USART_SendData(port, (uint8_t) c);
	
        /* Loop until the end of transmission */
        while (USART_GetFlagStatus(port, USART_FLAG_TC) == RESET);
}


int
owluart_inb(void *uart, unsigned int timeout)
{
        unsigned int ms = owltmr_get_ms();
        USART_TypeDef *port = uart;
        
        for (;;) {
                if (owluart_poll(port)) {
                        char c = USART_ReceiveData(port);
                        owl_dbg("0x%x %c", c, c); 
                        return c;
                }
                
                if (timeout > 0 &&
                    owltmr_intvl_passed(ms, owltmr_get_ms(), timeout)) {
                        owl_info("got timeout (%d)", timeout);
                        return OWL_ERR_TIMEOUT;
                }
        }
}

int
owluart_poll(void *uart)
{
        USART_TypeDef *port = uart;
        int overrun;

        overrun = USART_GetFlagStatus(port, USART_FLAG_ORE);
        if (overrun) {
                owl_err("overrun");
                (void) USART_ReceiveData(port);
        }
        
        return USART_GetFlagStatus(port, USART_FLAG_RXNE);
}


int
owluart_write(void *uart, const void *buf, int len)
{
        int i;
        const char *out = buf;

        for (i = 0; i < len; i++)
                owluart_outb(uart, out[i]);
        
        return len;
}
