/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */
 

#define OWL_TRACE OWL_TRACE_INFO
#define OWL_FILE "sys_arch"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_thread.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_err.h>

#include <owl/util/rmutex.h>

#include "lwip/sys.h"
#include "lwip/opt.h"
#include "lwip/stats.h"

#if !NO_SYS

struct sys_mbox_msg {
	struct sys_mbox_msg *next;
	void *msg;
};

#define SYS_MBOX_SIZE 128
struct sys_mbox {
	int first;
	int last;
	int count;
	void *msgs[SYS_MBOX_SIZE];
	struct sys_sem *post;
	struct sys_sem *fetch;
	struct sys_sem *mutex; 
};

struct sys_sem {
	owlsem_t sem;
};

static struct rmutex rmutex; /* for sys_arch_protect() */

static struct sys_sem *sys_sem_new_internal(u8_t count);
static void sys_sem_free_internal(struct sys_sem *sem);


sys_thread_t
sys_thread_new(const char *name, lwip_thread_fn function, void *arg,
	       int stacksize, int prio)
{
	return NULL;
}

err_t
sys_mbox_new(struct sys_mbox **mb, int size)
{
	struct sys_mbox *mbox;
	LWIP_UNUSED_ARG(size);

	mbox = (struct sys_mbox *) owl_os_zalloc(sizeof(struct sys_mbox));
	if (mbox == NULL) {
		return ERR_MEM;
	}
	mbox->post = sys_sem_new_internal(SYS_MBOX_SIZE);
	mbox->fetch = sys_sem_new_internal(0);
	mbox->mutex = sys_sem_new_internal(1);

	SYS_STATS_INC_USED(mbox);
	*mb = mbox;
	return ERR_OK;
}

void
sys_mbox_free(struct sys_mbox **mb)
{
	if ((mb != NULL) && (*mb != SYS_MBOX_NULL)) {
		struct sys_mbox *mbox = *mb;
		SYS_STATS_DEC(mbox.used);
		sys_sem_wait(&mbox->mutex);
		sys_sem_free_internal(mbox->fetch);
		sys_sem_free_internal(mbox->post);
		sys_sem_free_internal(mbox->mutex);
		owl_os_free(mbox);
	}
}

err_t
sys_mbox_trypost(struct sys_mbox **mb, void *msg)
{
	struct sys_mbox *mbox;
	LWIP_ASSERT("invalid mbox", (mb != NULL) && (*mb != NULL));
	mbox = *mb;

	if ((sys_sem_wait(&mbox->mutex)))
		return ERR_MEM;

	LWIP_DEBUGF(
		SYS_DEBUG, ("sys_mbox_trypost: mbox:%p msg:%p\n", mbox, msg));
	
	if (mbox->count == SYS_MBOX_SIZE) {
		sys_sem_signal(&mbox->mutex);
		LWIP_DEBUGF(
			SYS_DEBUG, ("sys_mbox_trypost: mbox:%p full\n", mbox));
		return ERR_MEM;
	}

	sys_sem_wait(&mbox->post); /* should never block */
	mbox->msgs[mbox->last] = msg;
	mbox->last = (mbox->last + 1) % SYS_MBOX_SIZE;
	mbox->count++;
	sys_sem_signal(&mbox->mutex);
	sys_sem_signal(&mbox->fetch);
	LWIP_DEBUGF(SYS_DEBUG, ("sys_mbox_trypost: mbox:%p done\n", mbox));
	return ERR_OK;
}


void
sys_mbox_post(struct sys_mbox **mb, void *msg)
{
	struct sys_mbox *mbox;
	LWIP_ASSERT("invalid mbox", (mb != NULL) && (*mb != NULL));
	mbox = *mb;

	/* Block until a slot is available */
	if ((sys_sem_wait(&mbox->post)))
		return;

	/* Grab the lock */
	if ((sys_sem_wait(&mbox->mutex)))
		return;

	LWIP_DEBUGF(SYS_DEBUG, ("sys_mbox_post: mbox:%p msg:%p\n", mbox, msg));

	mbox->msgs[mbox->last] = msg;
	mbox->last = (mbox->last + 1) % SYS_MBOX_SIZE;
	mbox->count++;
	sys_sem_signal(&mbox->mutex);
	sys_sem_signal(&mbox->fetch);
}


u32_t
sys_arch_mbox_tryfetch(struct sys_mbox **mb, void **msg)
{
	struct sys_mbox *mbox;
	LWIP_ASSERT("invalid mbox", (mb != NULL) && (*mb != NULL));
	mbox = *mb;

	if ((sys_sem_wait(&mbox->mutex))) {
		if (msg)
			*msg = NULL;
		return ERR_OK;
	}

	LWIP_DEBUGF(SYS_DEBUG, ("sys_mbox_tryfetch: mbox:%p\n", mbox));
	
	if (mbox->count == 0) {
		sys_sem_signal(&mbox->mutex);
		LWIP_DEBUGF(
			SYS_DEBUG,
			("sys_mbox_tryfetch: mbox:%p empty\n", mbox));
		
		return ERR_MEM;
	}

	/* should never block */
	sys_sem_wait(&mbox->fetch);

	if (msg != NULL)
		*msg = mbox->msgs[mbox->first];
	
	mbox->first = (mbox->first + 1) % SYS_MBOX_SIZE;
	mbox->count--;
	sys_sem_signal(&mbox->mutex); /* Might cause context switch */
	sys_sem_signal(&mbox->post);
	LWIP_DEBUGF(SYS_DEBUG, ("sys_mbox_tryfetch: mbox:%p done\n", mbox));
	return ERR_OK;
}

u32_t
sys_arch_mbox_fetch(struct sys_mbox **mb, void **msg, u32_t timeout)
{
	int err;
	uint32_t ms;
	struct sys_mbox *mbox;
	LWIP_ASSERT("invalid mbox", (mb != NULL) && (*mb != NULL));
	mbox = *mb;

	LWIP_DEBUGF(
		SYS_DEBUG,
		("sys_mbox_fetch: mbox:%p timeout:%d\n", mbox, timeout));
	
	/* Block until a message is available */
	ms = owltmr_get_ms();
	if ((err = sys_arch_sem_wait(
		     &mbox->fetch, timeout)) == SYS_ARCH_TIMEOUT) {
		
		uint32_t duration = owltmr_get_intvl(ms, owltmr_get_ms());
		
		LWIP_DEBUGF(
			SYS_DEBUG,
			("sys_mbox_fetch: timeout, mbox:%p duration:%d\n",
			 mbox, duration));
		
		if (duration < timeout)
			return duration;
		else
			return err;
	} else if (err) {
		/* Mbox was free'd */
		if (msg)
			*msg = NULL;
		return ERR_OK;
        }

	LWIP_DEBUGF(SYS_DEBUG, ("sys_mbox_fetch: mbox:%p wait lock\n", mbox));
	
	/* Grab the lock */
	if ((sys_sem_wait(&mbox->mutex))) {
		if (msg)
			*msg = NULL;
		return ERR_OK;
	}

	LWIP_DEBUGF(SYS_DEBUG, ("sys_mbox_fetch: mbox:%p got lock\n", mbox));
	
	if (msg != NULL)
		*msg = mbox->msgs[mbox->first];
	
	mbox->first = (mbox->first + 1) % SYS_MBOX_SIZE;
	mbox->count--;
	sys_sem_signal(&mbox->mutex);  /* Might cause context switch */
	sys_sem_signal(&mbox->post);
	LWIP_DEBUGF(SYS_DEBUG, ("sys_mbox_fetch: mbox:%p done\n", mbox));
	return ERR_OK;
}

static struct sys_sem *
sys_sem_new_internal(u8_t count)
{
	struct sys_sem *sem;

	sem = (struct sys_sem *) owl_os_alloc(sizeof(struct sys_sem));
	if (sem != NULL) {
		owlsem_init(&sem->sem, count);
	}
		
	return sem;
}

err_t
sys_sem_new(struct sys_sem **sem, u8_t count)
{
	SYS_STATS_INC_USED(sem);
	
	*sem = sys_sem_new_internal(count);
	if (*sem == NULL)
		return ERR_MEM;
	
	return ERR_OK;
}

u32_t
sys_arch_sem_wait(struct sys_sem **s, u32_t timeout)
{
	int err;
	struct sys_sem *sem = *s;
	
	LWIP_ASSERT("invalid sem", (s != NULL) && (*s != NULL));

	if ((err = owlsem_wait_tmo(&sem->sem, timeout)) == OWL_ERR_TIMEOUT)
		return SYS_ARCH_TIMEOUT;

	return err;
}

void
sys_sem_signal(struct sys_sem **s)
{
	struct sys_sem *sem;
	LWIP_ASSERT("invalid sem", (s != NULL) && (*s != NULL));
	sem = *s;
	owlsem_post(&sem->sem);
}

static void
sys_sem_free_internal(struct sys_sem *sem)
{
	owlsem_destroy(&sem->sem);
	owl_os_free(sem);
}

void
sys_sem_free(struct sys_sem **sem)
{
	if ((sem != NULL) && (*sem != SYS_SEM_NULL)) {
		SYS_STATS_DEC(sem.used);
		sys_sem_free_internal(*sem);
	}
}

#endif /* !NO_SYS */


u32_t
sys_now(void)
{
	return owltmr_get_ms();
}

void
sys_init(void)
{
	rmutex_init(&rmutex);
}

sys_prot_t
sys_arch_protect(void)
{
	return rmutex_lock(&rmutex);
}

void
sys_arch_unprotect(sys_prot_t pval)
{
	LWIP_UNUSED_ARG(pval);
	rmutex_unlock(&rmutex);
}

u32_t
sys_jiffies(void)
{
	return owltmr_get_ms();
}
