#include "lwip/opt.h"
#include "lwip/def.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/stats.h"
#include "lwip/sys.h"
#include "netif/etharp.h"
#include "netif/wlif.h"
#include <wl_api.h>
#include <wlap_api.h>

#define IFNAME0 'w'
#define IFNAME1 'l'

/* the queue size will affect the tx performance when using power save.
 * A small queue will quickly become filled up if we have to wake the device
 * before the actual transmission can occur. When the queue is filled up, the
 * packets will be discarded and retransmission will be handled by the upper
 * layers. In case of TCP, the retransmission time might be quite long.
 *
 * If the packets can be put in the pqueue instead, all the packets
 * (if possible) will be transmitted when the device wakes up, so we don't have
 * to wait for retransmission from upper layers.
 */
#ifndef WLIF_QUEUELEN
# define WLIF_QUEUELEN WL_TX_QUEUE_LENGTH
#endif

#ifdef WLIF_USE_POOLS
# if PBUF_POOL_SIZE < WLIF_QUEUELEN + 1
#  error "PBUF_POOL_SIZE < WLIF_QUEUELEN + 1"
# endif
#endif

struct wlif_t {
	volatile uint8_t rx_pending;

	struct {
		struct pbuf *buf[WLIF_QUEUELEN];
		uint8_t first;
		uint8_t last;
	} pqueue;
        
#if !NO_SYS
        sys_mbox_t mbox;
        sys_sem_t lock;
#endif /* !NO_SYS */
};

static void wlif_input(void *arg);

/* If NO_SYS == 0, make sure that WLANRX_TASK_PRIO is defined */

#define PQUEUE_EMPTY(q) (q.last == q.first)
#define PQUEUE_FULL(q) ((q.last + 1) % WLIF_QUEUELEN == q.first)
#define PQUEUE_FIRST(q) (q.buf[q.first])
#define PQUEUE_DEQUEUE(q)                                               \
        ({                                                              \
                struct pbuf* __p = PQUEUE_FIRST(q);                     \
                q.first = (q.first + 1) % WLIF_QUEUELEN;                  \
                __p;                                                    \
        })
#define PQUEUE_ENQUEUE(q, p)                                            \
        ({                                                              \
                q.buf[q.last] = p;                                      \
                q.last = (q.last + 1) % WLIF_QUEUELEN;                    \
        })


#if !NO_SYS
void
wlif_lock(void *arg)
{
        struct netif *netif = arg;
        struct wlif_t *priv = (struct wlif_t *) netif->state;

        sys_sem_wait(priv->lock);
}

void
wlif_unlock(void *arg)
{
        struct netif *netif = arg;
        struct wlif_t *priv = (struct wlif_t *) netif->state;

        sys_sem_signal(priv->lock);
}
#else
# define wlif_lock(...)
# define wlif_unlock(...)
#endif /* !NO_SYS */

static err_t process_pqueue(struct netif* netif)
{
        struct pbuf *p;
        struct pbuf *q;
        int status;
	struct wlif_t *priv = (struct wlif_t*) netif->state;

        wlif_lock(netif);

        /* queue empty? finished */
        if (PQUEUE_EMPTY(priv->pqueue)) {
                wlif_unlock(netif);
                return ERR_OK;
        }

        /* get first packet in queue */
        p = PQUEUE_FIRST(priv->pqueue);

        status = wl_process_tx(
                p->payload + WL_HEADER_SIZE, /* ptr to eth hdr */
                p->len - WL_HEADER_SIZE,     /* input buffer len */
                p->tot_len - WL_HEADER_SIZE, /* pkt len */
                p->payload,                  /* ptr to WE hdr */
                0,                           /* prio */
                p);                          /* pkt handle */


        /* if we fail due to power save mode, leave packet in queue and
         * try again when device is awake again.
         */
	if (status == WL_RESOURCES) {
                wlif_unlock(netif);
                LWIP_DEBUGF(NETIF_DEBUG, ("process_tx fail WL_RESOURCES\n"));
                return ERR_IF;
        }

        /* if we fail for another reason, just discard the packet */
        if (status != WL_SUCCESS) {
                PQUEUE_DEQUEUE(priv->pqueue);
                wlif_unlock(netif);
                LWIP_DEBUGF(NETIF_DEBUG, ("process_tx fail:%d\n", status));
                pbuf_free(p);
                return ERR_IF;
        }

        /* Send the data from the pbuf to the interface, one pbuf at a
         * time. The size of the data in each pbuf is kept in the ->len
         * variable.
         */
        for (q = p; q != NULL; q = q->next)
		wl_tx(q->payload, q->len);

        /* remove packet from queue and dec refcnt */
        PQUEUE_DEQUEUE(priv->pqueue);
        wlif_unlock(netif);
        pbuf_free(p);
        
        LINK_STATS_INC(link.xmit);

        /* tell caller to process next packet */
        return ERR_INPROGRESS;
}


/**
 * Called in interrupt context when we can read more data from the mac.
 *
 */
static void
rx_isr(void* ctx)
{
	struct netif *netif = ctx;
	struct wlif_t *priv = (struct wlif_t*) netif->state;
#if !NO_SYS
        void *msg = NULL;
#endif /* !NO_SYS */
        
        priv->rx_pending = 1;

#if !NO_SYS
        /* NOTE: it must be valid to call sys_mbox_trypost() from an isr */
        if (sys_mbox_trypost(priv->mbox, &msg) != ERR_OK) {
                /* dont print anything in isr */
                return;
        }
#endif /* !NO_SYS */
}


static ssize_t
pkt_read_cb(char *dst, void *src_handle, size_t read_len, int offset)
{
        ssize_t rc;

        if ((rc = pbuf_copy_partial((struct pbuf *)src_handle, dst, read_len,
                                    offset + WL_HEADER_SIZE)) == 0)
                return -1;

        return rc;
}


/**
 * In this function, the hardware should be initialized.
 * Called from wlif_init().
 *
 * @param netif the already initialized lwip network interface structure
 *        for this ethernetif
 */
static err_t
low_level_init(struct netif *netif)
{
#if !NO_SYS
        struct wlif_t* priv = (struct wlif_t*) netif->state;
#endif /* !NO_SYS */
         
        /* device capabilities */
        netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP |
                NETIF_FLAG_IGMP;
        /* NETIF_FLAG_LINK_UP must be set only when we have an wlan assoc */

        /* set MAC hardware address length */
        netif->hwaddr_len = ETHARP_HWADDR_LEN;

	if (wl_get_mac_addr(netif->hwaddr) != WL_SUCCESS) {
                LWIP_DEBUGF(NETIF_DEBUG, ("fail to get mac address\n"));
		return ERR_IF;
        }

        /* maximum transfer unit */
        netif->mtu = 1500;

#if !NO_SYS
        LWIP_DEBUGF(NETIF_DEBUG, ("create mbox\n"));
        if ((priv->mbox = sys_mbox_new(1)) == SYS_MBOX_NULL) {
                LWIP_DEBUGF(NETIF_DEBUG, ("fail to create mbox\n"));
		return ERR_IF;
        }

        LWIP_DEBUGF(NETIF_DEBUG, ("create lock sem\n"));
        if ((priv->lock = sys_sem_new(1)) == SYS_SEM_NULL) {
                LWIP_DEBUGF(NETIF_DEBUG, ("fail to create sem\n"));
		return ERR_IF;
        }

        LWIP_DEBUGF(NETIF_DEBUG, ("create rx thread\n"));
	if (sys_thread_new("wlanrx", wlif_input, netif, 4096,
                           WLANRX_TASK_PRIO) == NULL) {
                LWIP_DEBUGF(NETIF_DEBUG, ("fail to create rx thread\n"));
		return ERR_IF;
        }
#endif /* !NO_SYS */

	wl_register_rx_isr(rx_isr, netif);
        wl_register_pkt_read_cb(pkt_read_cb);
        return ERR_OK;
}


/**
 * This function should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @param p the MAC packet to send (e.g. IP packet including MAC addresses and
 *        type)
 * @return ERR_OK if the packet could be sent
 *         an err_t value if the packet couldn't be sent
 *
 * @note Returning ERR_MEM here if a DMA queue of your MAC is full can lead to
 *       strange results. You might consider waiting for space in the DMA queue
 *       to become availale since the stack doesn't retry to send a packet
 *       dropped because of memory failure (except for the TCP timers).
 */
static err_t
low_level_output(struct netif *netif, struct pbuf *p)
{  
	struct wlif_t* priv = (struct wlif_t*) netif->state;

        /* must have a linear buffer containing up to and including
         * the ethernet header
         */
        if (p->len < 14)
                return ERR_IF;


        wlif_lock(netif);
        
        /* queue full? drop packet */
        if (PQUEUE_FULL(priv->pqueue))  {
                wlif_unlock(netif);
                LWIP_DEBUGF(NETIF_DEBUG, ("queue full, drop packet"));
                return ERR_INPROGRESS; /* no one seems to check this anyway */
        }
		
        /* queue packet */
        PQUEUE_ENQUEUE(priv->pqueue, p);
        wlif_unlock(netif);
        pbuf_ref(p);
        while (process_pqueue(netif) == ERR_INPROGRESS);
        return ERR_OK; /* no one seems to check this anyway */
}

/**
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return a pbuf filled with the received packet (including MAC header)
 *         NULL on memory error
 */
static struct pbuf *
low_level_input(struct netif *netif)
{
        struct pbuf *p;
	struct wlif_t *priv = (struct wlif_t*) netif->state;

        char *stripped_pkt;
        size_t stripped_pkt_len;
        u16_t vlan;
        u8_t rx_hdr_size;
        int status;
        u16_t len;

        /* maximum packet length from wl_rx() */
        len = WL_MAX_PKT_LEN;

        /* We allocate a continous pbuf */
        while ((p = pbuf_alloc(PBUF_RAW, len,
#ifdef WLIF_USE_POOLS
                               PBUF_POOL
#else
                               PBUF_RAM
#endif
                        )) == NULL) {
                LWIP_DEBUGF(NETIF_DEBUG, ("low_level_input: fail to alloc "
                                          "pbuf of len:%"S32_F"\n", len));
#if !NO_SYS
                /* let other threads run to free up some memory. There is no
                   interface in sys_arch to do yield/sleep so this must be
                   modified for the specific rtos used */
                wlif_yield();
#else
                return NULL;
#endif /* !NO_SYS */
        }

        /* Read the entire msg */
	priv->rx_pending = 0;
        
        wlif_lock(netif);
        
        wl_rx(p->payload, &len);
        if (len == 0) {
                wlif_unlock(netif);
                LWIP_DEBUGF(NETIF_DEBUG, ("low_level_input: len was 0\n"));
                return NULL;
        }

        status = wl_process_rx(
                p->payload,             /* input buf */
                len,                    /* input buf length */
                &stripped_pkt,          
                &stripped_pkt_len,      
                &vlan);

        wlif_unlock(netif);

        if (status == WL_ABSORBED) {
                LWIP_DEBUGF(NETIF_DEBUG, ("low_level_input: absorbed\n"));
                pbuf_free(p);
                return NULL;
        }
		
        /* Data packet, remove padding */
        rx_hdr_size = stripped_pkt - (char*) p->payload;
        pbuf_realloc(p, stripped_pkt_len + rx_hdr_size);
        
        LINK_STATS_INC(link.recv);
        LWIP_DEBUGF(NETIF_DEBUG, ("low_level_input: packet to lwip\n"));
        return p;  
}


/**
 * This function will be called by wlif_poll() when a packet has been received
 * from the mac. Then the type of the received packet is determined and
 * the appropriate input function is called.
 *
 * @param netif the lwip network interface structure for this ethernetif
 */
static void
wlif_input(void *arg)
{
        struct netif *netif = arg;
        struct eth_hdr *ethhdr;
        struct pbuf *p;

#if !NO_SYS
        struct wlif_t *priv = (struct wlif_t *) netif->state;
        void *msg;
        
	for (;;) {
                /* wait until signal from rx isr */
                sys_mbox_fetch(priv->mbox, &msg);
                
#endif /* !NO_SYS */

                /* move received packet into a new pbuf */
                p = low_level_input(netif);
                
                /* no packet could be read or the packet was absorbed,
                   silently ignore this */
                if (p == NULL) {
                
                        /* Send any packets that was queued due to filled up
                           target queue or power save mode to free up memory */
                        process_pqueue(netif);

#if !NO_SYS
                        continue;
#else
                        return;
#endif /* !NO_SYS */
                }
                        
                /* points to packet payload, which starts with an Ethernet
                   header */
                ethhdr = p->payload;
                switch (htons(ethhdr->type)) {
                        /* IP or ARP packet? */
                case ETHTYPE_IP:
                case ETHTYPE_ARP:
#if PPPOE_SUPPORT
                        /* PPPoE packet? */
                case ETHTYPE_PPPOEDISC:
                case ETHTYPE_PPPOE:
#endif /* PPPOE_SUPPORT */
                        LWIP_DEBUGF(NETIF_DEBUG, ("wlif_input: to lwip\n"));
                        /* full packet send to tcpip_thread to process */
                        if (netif->input(p, netif) != ERR_OK) {
                                LWIP_DEBUGF(NETIF_DEBUG,
                                            ("wlif_input: IP input error\n"));
                                pbuf_free(p);
                                p = NULL;
                        }
                        break;

                default:
                        LWIP_DEBUGF(NETIF_DEBUG, ("wlif_input: inval type\n"));
                        pbuf_free(p);
                        p = NULL;
                        break;
                }

                /* Send any packets that was queued due to filled up
                   target queue or power save mode */
                process_pqueue(netif);

#if !NO_SYS
        }
#endif /* !NO_SYS */
}

/**
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 * This function should be passed as a parameter to netif_add().
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return ERR_OK if the loopif is initialized
 *         ERR_MEM if private data couldn't be allocated
 *         any other err_t on error
 */
err_t
wlif_init(struct netif *netif)
{
        static struct wlif_t wlif;

        LWIP_ASSERT("netif != NULL", (netif != NULL));

#if LWIP_NETIF_HOSTNAME
        /* Initialize interface hostname */
        if ( NULL == netif->hostname ) {
                netif->hostname = "wlif";
        }
#endif /* LWIP_NETIF_HOSTNAME */

	netif->state = &wlif;
        netif->name[0] = IFNAME0;
        netif->name[1] = IFNAME1;

        /* We directly use etharp_output() here to save a function call.
         * You can instead declare your own function an call etharp_output()
         * from it if you have to do some checks before sending (e.g. if link
         * is available...) */
        netif->output = etharp_output;
        netif->linkoutput = low_level_output;

        /* initialize the hardware */
        return low_level_init(netif);
}

/**
 * When the host enters sleep modes and is woken by interrupts
 * from the WiFi it is useful to know if an interrupt is pending
 * before going back to sleep.
 */
int
wlif_is_rx_pending(struct netif *netif)
{
        struct wlif_t* priv = NULL;

        if (netif)
                priv = (struct wlif_t*) netif->state;
        return priv->rx_pending;
}

void
wlif_reinit(struct netif *netif)
{
}


#if NO_SYS
/**
 *
 */
void
wlif_poll(struct netif* netif) 
{
        struct wlif_t* priv = NULL;

        /* wl api forward progress */
        wl_poll();

        if (netif)
                priv = (struct wlif_t*) netif->state;

        /* wlif_init() not called yet? */
        if (priv == NULL)
                return;

	/* no packets pending? */
	if (!priv->rx_pending)
		return;

	/* read the pending packet and process the packet queue */
	wlif_input(netif);
}
#endif /* NO_SYS */
