#define OWL_TRACE OWL_TRACE_ERR

#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <wl_os.h>

#include "lwip/opt.h"

#include "lwip/mem.h"
#include "lwip/raw.h"
#include "lwip/icmp.h"
#include "lwip/netif.h"
#include "lwip/sys.h"
#include "lwip/sockets.h"
#include "lwip/inet.h"
#include "lwip/inet_chksum.h"
#include "lwip/tcp.h"

#include <lwip/util/tcp_wrap.h>

struct tcp_handle {
        struct tcp_pcb *lpcb; /* Listen PCB */
        struct tcp_pcb *dpcb; /* Data PCB     */

        /* Connection config */
        void (*conn_monitor_cb)(tcp_wrap_handle_t h, int status, void *ctx);
        void *conn_monitor_ctx;
        void (*listen_connected_cb)(void *ctx);
        void *listen_connected_ctx;

        /* Rx config */
        void (*rx_cb)(const void *data, size_t len, void *ctx);
        void *rx_cb_ctx;
        uint8_t tx_ready;

        /* Tx config */
        void (*tx_complete_cb)(int status, const void *data, void *ctx);
        void *tx_ctx;
        const char *tx_data;
        size_t tx_data_len;
        size_t tx_data_off;

        /* To allow free from within callbacks */
        uint8_t free_status;
};

enum {
        FREE_BLOCKED = 1,
        FREE_PENDING = 2
};

static void _data_close(tcp_wrap_handle_t handle);
static void _do_tx(tcp_wrap_handle_t handle);

static err_t _rx_cb(void *ctx,
                    struct tcp_pcb *pcb,
                    struct pbuf *pbuf,
                    err_t err) {
        struct tcp_handle *h = ctx;
        uint16_t rx_len = 0;

        owl_dbg(" Entry \n");
        if ( NULL == pbuf ) {
                owl_dbg(" Connection closed by peer\n");
                _data_close(h);
                return err;
        }

        owl_dbg(" Rx %d bytes\n", pbuf->tot_len);
        if ( err != ERR_OK ) {
                owl_dbg(" got error code %d\n", err);
                _data_close(h);
                return err;
        }
        if ( NULL == h->rx_cb ) {
                /* No receiver registered, sink the data */
                owl_dbg(" sinking data\n");
                rx_len = pbuf->tot_len;
                goto out;
        }

        if ( h->rx_cb ) {
                struct pbuf *iter;
                owl_dbg("pbuf %p, tot_len %d, rx_handler %p", pbuf, pbuf->tot_len,
                        h->rx_cb);
                iter = pbuf;
                while (iter) {
                        owl_dbg("iter %d, %p", iter->len, iter->next);
                        h->rx_cb(iter->payload,
                                 iter->len,
                                 h->rx_cb_ctx);
                        rx_len += iter->len;
                        iter = iter->next;
                }
        }
out:
        owl_dbg(" acking %d bytes\n", rx_len);
        tcp_recved(pcb, rx_len);
        pbuf_free(pbuf);
        return ERR_OK;
}

static err_t _tx_done_cb(void *ctx,
                         struct tcp_pcb *pcb,
                         uint16_t len) {
        struct tcp_handle *h = ctx;

        (void) pcb;

        owl_dbg(" Entry (len %d (remains %d bytes))\n",
                len,
                h->tx_data_len - len);

        /* Good write, update the offsets */
        if ( h->tx_data_len ) {
                h->tx_data_off += len;
                h->tx_data_len -= len;
        }

        _do_tx(h);

        return ERR_OK;
}

static void _do_tx(tcp_wrap_handle_t handle) {
        size_t chunk_len;
        struct tcp_handle *h = handle;
        err_t err;

        owl_dbg(" Entry \n");
        if ( ! h->tx_ready || NULL == h->tx_data ) {
                owl_dbg(" tx_ready %d, h->tx_data %p\n",
                        h->tx_ready, h->tx_data);
                return;
        }
        if ( 0 == h->tx_data_len ) {
                /* Do we have to call the completion callback? */
                if ( h->tx_data && h->tx_complete_cb ) {
                        const char *data = h->tx_data;
                        owl_dbg(" calling tx complete\n");
                        h->tx_data = NULL;
                        h->tx_complete_cb( 1, data, h->tx_ctx );
                }
                return;
        }
        chunk_len = tcp_sndbuf(h->dpcb);

        if ( h->tx_data_len - h->tx_data_off < chunk_len ) {
                chunk_len = h->tx_data_len - h->tx_data_off;
        }
        owl_dbg(" Sending %d byte chunk from buf %p+%X\n",
                chunk_len,
                h->tx_data,
                h->tx_data_off);
        err = tcp_write(h->dpcb,
                        h->tx_data + h->tx_data_off,
                        chunk_len,
                        0 /* no copy */);
        if ( ERR_MEM == err ) {
                owl_dbg(" Out of memory. Postponing tx\n");
                return;
        }
        if ( err != ERR_OK ) {
                owl_dbg(" Failed to send data. Closing connection\n");
                _data_close(h);
                return;
        }
        if (chunk_len > 32)
                tcp_output(h->dpcb);
}


/* Called when a connection initiated by us is completed
 * (we're the client).
 */
static err_t _connected_cb(void *ctx,
                          struct tcp_pcb *pcb,
                          err_t err) {
        struct tcp_handle *h = ctx;
        int status = 1;

        (void) pcb;

        owl_dbg(" Connected (handle %p, dpcb %p)\n",
                h,
                h->dpcb);
        if ( err != ERR_OK ) {
                status = 0; /* Failed */
                owl_dbg(" Error status %d\n", err);
                goto out;
        }
        tcp_sent(h->dpcb,
                 _tx_done_cb);

        h->tx_ready = 1;
        _do_tx(h);

out:
        if ( h->conn_monitor_cb ) {
                h->conn_monitor_cb((tcp_wrap_handle_t)h,
                                   status,
                                   h->conn_monitor_ctx);
        }

        return ERR_OK;
}

static void _conn_err_cb(void *ctx, err_t err) {
        struct tcp_handle *h = ctx;

        (void) err;

        owl_dbg(" Entry, error code %d\n", err);

        /* If the callback frees the handle we need
         * to postpone the actual freeing */
        h->free_status = FREE_BLOCKED;

        h->dpcb = NULL;
        if ( h->rx_cb ) {
                owl_dbg("%s rx_cb()\n");
                h->rx_cb(NULL, 0, h->rx_cb_ctx);
        }
        if ( h->tx_data && h->tx_complete_cb ) {
                owl_dbg("%s tx_complete_cb()\n");
                h->tx_complete_cb( 0, h->tx_data, h->tx_ctx );
        }
        if ( h->conn_monitor_cb ) {
                owl_dbg("%s conn_monitor_cb()\n");
                h->conn_monitor_cb((tcp_wrap_handle_t)h,
                                   0,
                                   h->conn_monitor_ctx);
        }
        h->tx_ready = 0;
        h->tx_data = NULL;
        h->tx_data_len = 0;
        h->tx_data_off = 0;
        if (FREE_PENDING == h->free_status) {
                h->free_status = 0;
                tcp_wrap_free(h);
        }
}

/* This function is called when a connection to us has
 * been accepted (we're the server).
 */
static err_t _accept_conn_cb(void *ctx,
                             struct tcp_pcb *pcb,
                             err_t err) {
        struct tcp_handle *h = ctx;

        owl_dbg("Accepted connection on handle %p, pcb %p\n",
                h,
                pcb);
        if ( err != ERR_OK ) {
                return err;
        }
        h->dpcb = pcb;
        h->tx_ready = 1;
        tcp_accepted(pcb);
        tcp_recv(h->dpcb, _rx_cb);
        tcp_sent(h->dpcb, _tx_done_cb);
        tcp_err(h->dpcb, _conn_err_cb);
        if ( h->listen_connected_cb ) {
                h->listen_connected_cb(h->listen_connected_ctx);
        }
        if ( h->conn_monitor_cb ) {
                owl_dbg("%s conn_monitor_cb()\n");
                h->conn_monitor_cb((tcp_wrap_handle_t)h,
                                   1,
                                   h->conn_monitor_ctx);
        }
        _do_tx(h);

        return ERR_OK;
}

/* Close the data PCBs (leaving the any listen PCB active)
 */
static void _data_close(tcp_wrap_handle_t handle) {
        struct tcp_handle *h = handle;
        err_t ret;

        if (h->dpcb) {
                owl_dbg("Closing dpcb\n");
                ret = tcp_close(h->dpcb);
                if ( ret != ERR_OK ) {
                        tcp_abort(h->dpcb);
                }
                if ( h->rx_cb ) {
                        h->rx_cb(NULL, 0, h->rx_cb_ctx);
                }
                h->tx_ready = 0;
                if ( h->tx_data && h->tx_complete_cb ) {
                        h->tx_complete_cb( 0, h->tx_data, h->tx_ctx );
                        h->tx_data = NULL;
                        h->tx_data_len = 0;
                        h->tx_data_off = 0;
                }
                h->dpcb = NULL;
        }
        if ( h->conn_monitor_cb ) {
                owl_dbg("%s conn_monitor_cb()\n");
                h->conn_monitor_cb((tcp_wrap_handle_t)h,
                                   0,
                                   h->conn_monitor_ctx);
        }
}

void tcp_wrap_listen(tcp_wrap_handle_t handle,
                     uint16_t port,
                     void (*connect_complete_cb)(void *ctx),
                     void *cb_ctx) {
        struct tcp_handle *h = handle;
        void *p;
        err_t ret;

        owl_dbg(" Entry\n");
        if ( NULL == h->lpcb ) {
                h->lpcb = tcp_new();
                tcp_arg(h->lpcb, h);
        }
        ret = tcp_bind(h->lpcb, IP_ADDR_ANY, port);
        owl_assert(ret == ERR_OK);
        if (ret != ERR_OK) {
                owl_info("err is %d");
        }
        p = tcp_listen(h->lpcb);
        owl_assert(p);
        owl_dbg("listening\n");
        h->lpcb = p;
        h->listen_connected_cb = connect_complete_cb;
        h->listen_connected_ctx = cb_ctx;
        tcp_accept(h->lpcb, _accept_conn_cb);
        owl_dbg("Accepted (handle %p)\n", h);
        owl_dbg("done\n");
}

void tcp_wrap_close(tcp_wrap_handle_t handle) {
        struct tcp_handle *h = handle;
        err_t ret;

        owl_dbg(" Entry (handle %p)\n", handle);
        _data_close(h);
        owl_dbg(" Entry\n");
        if (h->lpcb) {
                ret = tcp_close(h->lpcb);
                if ( ret != ERR_OK ) {
                        tcp_abort(h->lpcb);
                }
                h->lpcb = NULL;
        }
}

void tcp_wrap_set_rx_handler(tcp_wrap_handle_t handle,
                             void (*rx_handler)(const void *data,
                                                size_t len,
                                                void *ctx),
                             void *ctx) {
        struct tcp_handle *h = handle;

        owl_dbg(" Entry (%p)\n", rx_handler);
        h->rx_cb = rx_handler;
        h->rx_cb_ctx = ctx;
}


void tcp_wrap_monitor_conn(tcp_wrap_handle_t handle,
                           void (*conn_monitor_cb)(tcp_wrap_handle_t handle,
                                                   int status,
                                                   void *ctx),
                      void *cb_ctx) {
        struct tcp_handle *h = handle;
        h->conn_monitor_cb = conn_monitor_cb;
        h->conn_monitor_ctx = cb_ctx;
}

int tcp_wrap_connect(tcp_wrap_handle_t handle,
                     struct ip_addr addr,
                     uint16_t port) {
        struct tcp_handle *h = handle;
        err_t ret;

        owl_dbg(" Entry\n");
        if ( h->dpcb || h->lpcb ) {
                owl_dbg(" Already in server mode. Failed.\n");
                return -1;
        }
        h->dpcb = tcp_new();
        owl_dbg(" Created dpcb %p\n", h->dpcb);
        tcp_arg(h->dpcb, h);
        tcp_err(h->dpcb, _conn_err_cb);
        tcp_recv(h->dpcb, _rx_cb);
        ret = tcp_connect(h->dpcb,
                          &addr,
                          port,
                          _connected_cb);
        if ( ret != ERR_OK ) {
                owl_dbg(" Failed to connect (error %d)\n", ret);
                return -1;
        }

        return 1;
}

/* Return < 0 on failure.
 * 0 to try again later.
 * > 0 on success.
 * data buffer may not be freed until tx_complete_cb is called.
 */
int tcp_wrap_tx(tcp_wrap_handle_t handle,
                const void *data,
                size_t len,
                void (*tx_complete_cb)(int status,
                                       const void *data,
                                       void *ctx),
                void *cb_ctx) {

        struct tcp_handle *h = handle;

        owl_dbg(" Entry (%d bytes@%p)\n", len, data);
        if ( h->tx_data ) {
                owl_dbg("Tx already pending (%d bytes).\n",
                        h->tx_data_len);
                return 0;
        }
        if ( NULL == h->dpcb || len > UINT16_MAX) {
                return -1;
        }
        if ( ! h->tx_ready ) {
                owl_dbg(" Not connected\n");
                return -1;
        }
        h->tx_complete_cb = tx_complete_cb;
        h->tx_ctx      = cb_ctx;
        h->tx_data     = data;
        h->tx_data_len = len;
        h->tx_data_off = 0;

        _do_tx(h);

        return 1;
}

int tcp_wrap_is_connected(tcp_wrap_handle_t handle) {
        struct tcp_handle *h = handle;

        return h->tx_ready;
}

/*!
 * Initialize a TCP wrapper handle.
 * @return A TCP wrapper handle.
 */
tcp_wrap_handle_t tcp_wrap_alloc(void) {
        struct tcp_handle *h;

        h = owl_os_zalloc(sizeof *h);
        owl_dbg(" Entry (%p)\n", h);

        return h;
}

void tcp_wrap_free(tcp_wrap_handle_t handle) {
        struct tcp_handle *h = handle;

        owl_dbg(" Entry\n");
        tcp_wrap_close(h);
        if ( h->free_status != 0 ) {
                owl_dbg(" Postpone free of tcp_wrap handle %p\n", h);
                h->free_status = FREE_PENDING;
                return;
        }
        owl_dbg(" Free tcp wrap handle %p\n", h);
        owl_os_free(h);
}

