#ifndef _TCP_WRAP_H
#define _TCP_WRAP_H

typedef void * tcp_wrap_handle_t;

/*! 
 * Register a rx handler.
 * The rx_handler callback will be called with received data as
 * it arrives.
 * The data parameter will be automatically freed after rx_handler
 * returns. If the data parameter is NULL that means that the connection
 * was closed.
 * @param handle tcp_wrapper handle.
 * @param rx_handler rx data handler function to register.
 * @param rx_ctx Opaque context that will be passed to the rx_handler
 *        function when it is called.
 */
void tcp_wrap_set_rx_handler(tcp_wrap_handle_t handle_t, 
                             void (*rx_handler)(const void *data, 
                                                size_t len, 
                                                void *ctx),
                             void *rx_ctx);

/*! Start to listen for connections on a tcp handle.
 * The handle must be closed (by calling tcp_wrap_close() 
 * before calling this function on it again.
 * @param handle tcp_wrapper handle.
 * @param port TCP port to listen to.
 * @param connect_complete_cb Callback to get connection
 *        established notifications. It is called each time
 *        a new TCP connection has been accepted on the
 *        listening socket. Can be NULL. This callback is
 *        not called if an established connection is closed,
 *        see \a tcp_wrap_monitor_connection() for that.
 * @param cb_ctx Opaque context that will be passed to the
 *        connect_complete_cb function when it is called.
 */
void tcp_wrap_listen(tcp_wrap_handle_t handle, 
                     uint16_t port,
                     void (*connect_complete_cb)(void *ctx),
                     void *cb_ctx);


/*! Send a data buffer.
 * The handle must be connected.
 * The data buffer must be valid until the tx_complete_cb() callback
 * is called with the data buffer as the data parameter (after which
 * the data buffer may be freed). The status parameter is 1 if the
 * data was successfully transmitted and 0 if the connection was
 * terminated before all the data could be sent.
 * @param handle tcp_wrapper handle.
 * @param data Data buffer to send.
 * @param len Length of data buffer in bytes.
 * @param tx_complete_cb Callback to call when the data has been
 *        sent or when the transfer is aborted.
 * @param cb_ctx Opaque context that will be passed to the
 *        tx_complete_cb function when it is called.
 * @returns
 * - < 0 on failure.
 * - 0 to try again later.
 * - > 0 on success.
 */
int tcp_wrap_tx(tcp_wrap_handle_t handle,
                const void *data,
                size_t len,
                void (*tx_complete_cb)(int status, 
                                       const void *data, 
                                       void *ctx),
                void *cb_ctx);

/*!
 * Connect to a remote server.
 * @param handle tcp_wrapper handle.
 * @param addr IP address to connect to.
 * @param port Port to connect to.
 * @return 
 *  - < 0 on failure.
 *  - 1 on success.
 */
int tcp_wrap_connect(tcp_wrap_handle_t handle,
                     struct ip_addr addr,
                     uint16_t port);

/*!
 * Monitor a connection. The callback is registered with the
 * tcp_wrapper handle until deleted by the user. This overwrites the
 * current monitoring callback. Call with conn_mon_cb NULL to stop
 * monitoring.
 * @param handle tcp_wrapper handle.
 * @param conn_mon_cb Callback that is called when a connection
 *        is completed or fails.
 *        The status parameter will be 1 on a successful connection 
 *        and 0 on failure. If this callback is called with status 1
 *        it will be called again when the established connection is 
 *        terminated (with status 0).
 * @param cb_ctx Opaque context that will be passed to the 
 *        connect_complete_cb function when it is called.
 * @param connect_monitor_cb Callback that is called when a
 *        connection is disconnected.
 */
void tcp_wrap_monitor_conn(tcp_wrap_handle_t handle,
                           void (*connect_monitor_cb)(tcp_wrap_handle_t handle, 
                                                      int status, 
                                                      void *ctx),
                           void *cb_ctx);
/*!
 * Close a tcp_wrapper handle.
 * The handle still has to be freed through a call to 
 * \a tcp_wrap_free().
 * @param handle Handle to close.
 */
void tcp_wrap_close(tcp_wrap_handle_t handle);

/*! Allocate a new tcp_wrapper handle.
 * @return A allocated tcp_wrapper handle.
 */
tcp_wrap_handle_t tcp_wrap_alloc(void);

/*! Free a tcp_wrapper handle.
 * @param handle Handle to free.
 */
void tcp_wrap_free(tcp_wrap_handle_t handle);

/*! Check if a tcp_wrapper handle is connected.
 * @param handle Handle to check.
 * @return 1 if the handle is connected (ready to accept
 *         transmissions). 0 otherwise.
 */
int tcp_wrap_is_connected(tcp_wrap_handle_t handle);


/* These are some adaptations for platforms with
 * real, blocking BSD sockets.
 */

#ifdef FEAT_SOCKETS

#ifdef __WXWINDOWS__ /* wxWindows sockets, tcp_wrap_wx.c */

 typedef wxSocketBase* sock_t;

#else /* BSD sockets, tcp_wrap_sock.c */

 typedef int sock_t;

#endif

 void tcp_wrap_rx(tcp_wrap_handle_t handle);

 void tcp_wrap_set_dsock(tcp_wrap_handle_t handle, sock_t sock);

 sock_t tcp_wrap_get_dsock(tcp_wrap_handle_t handle);

#endif /* FEAT_SOCKETS */


#endif /* _TCP_WRAP_H */
