/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/drivers/owl_board.h>

/* This file uses the following board configuration defines:
 *
 * OWL_BOARD_HDG104_POWER_PIN         Power to hdg104, connected through
 *                                      transistor which inverts the
 *                                      logic, so set the pin from the avr to
 *                                      "high" to power-off the wifi. Optional.
 * OWL_BOARD_HDG104_ACTIVE_PIN        Output from reset-circuit to avr32.
 *                                      Optional.
 * OWL_BOARD_HDG104_SHUTDOWN_PIN      Shutdown signal connected directly to
 *                                      hdg104. Active low.
 *                                      Only defined if the irq is used on the
 *                                      board.
 */

int
hdg104_resume(void)
{
#if defined(OWL_BOARD_HDG104_ACTIVE_PIN)
        int shutdown_ms = 0;
        const int shutdown_tmo = 350;
#endif
        
        /* Configure power-pins */
        
#if defined(OWL_BOARD_HDG104_POWER_PIN)                 
        /* power to hdg104, connected through transistor which inverts the
           logic, so set the pin from the avr to "high" to power-off the wifi */
        owlgpio_cfg(OWL_BOARD_HDG104_POWER_PIN, OWL_GPIO_OUTPUT);
        owlgpio_set(OWL_BOARD_HDG104_POWER_PIN); /* power-off */
#endif

#if defined(OWL_BOARD_HDG104_SHUTDOWN_PIN)
        /* Shutdown signal to connected directly to hdg104. Active low */
        owlgpio_cfg(OWL_BOARD_HDG104_SHUTDOWN_PIN, OWL_GPIO_OUTPUT);
        owlgpio_clr(OWL_BOARD_HDG104_SHUTDOWN_PIN); /* shutdown wifi */
        
#elif defined (OWL_BOARD_HDG104_ACTIVE_PIN)
        /* Output from reset-circuit to avr32. This signal will indicate when
           the shutdown signal (generated by reset-circuit) is stable "high",
           e.g. the wifi is activated. We will wait for this pin to go high. */
        owlgpio_cfg(OWL_BOARD_HDG104_ACTIVE_PIN, OWL_GPIO_INPUT);
#endif /* HDG104_SHUTDOWN_xxx_PIN */


        /* Power-on device */
        
#if defined(OWL_BOARD_HDG104_POWER_PIN)
        /* enable main power supply to hdg104 */
        owlgpio_clr(OWL_BOARD_HDG104_POWER_PIN);
#endif

        owlboard_mdelay(5);

#if defined(OWL_BOARD_HDG104_SHUTDOWN_PIN)
        /* we don't have a reset circuit, so set shutdown 5 ms after power is
           enabled */
        owlgpio_set(OWL_BOARD_HDG104_SHUTDOWN_PIN);
        
#elif defined (OWL_BOARD_HDG104_ACTIVE_PIN)
        /* The reset circuit will set the shutdown pin once the power reached
           a proper level. The ouput from the reset circuit will go high when
           this occur */
        while (owlgpio_is_low(OWL_BOARD_HDG104_ACTIVE_PIN)) {
                if (shutdown_ms > shutdown_tmo)
                        return OWL_ERR_TIMEOUT;
                
                owlboard_mdelay(5);
                shutdown_ms += 5;
        }
#endif /* HDG104_SHUTDOWN_xxx_PIN */

        return 0;
}

int
hdg104_suspend(void)
{
#if defined(OWL_BOARD_HDG104_POWER_PIN)                 
        /* power to hdg104, connected through transistor which inverts the
           logic, so set the pin from the avr to "high" to power-off the wifi */
        owlgpio_cfg(OWL_BOARD_HDG104_POWER_PIN, OWL_GPIO_OUTPUT);
        owlgpio_set(OWL_BOARD_HDG104_POWER_PIN); /* power-off */
#endif
    
#if defined(OWL_BOARD_HDG104_SHUTDOWN_PIN)
        /* Shutdown signal to connected directly to hdg104. Active low */
        owlgpio_cfg(OWL_BOARD_HDG104_SHUTDOWN_PIN, OWL_GPIO_OUTPUT);
        owlgpio_clr(OWL_BOARD_HDG104_SHUTDOWN_PIN); /* shutdown wifi */
#endif
        return 0;
}
