/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

#include <wlap_api.h>

#if defined(OWL_CONFIG_HOSTAPD)
# include <hostapd/hostapd_api.h>
#endif

#include <owl/core/owl_ap.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>

struct owlap_ctx {
        owlap_link_cb_t *link_cb;
        void *link_ctx;
        int running;
        volatile int init_completed;
};


static struct owlap_ctx *priv = NULL;

static void event_cb(struct wl_event_t event, void *ctx);
static void wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx);
static void init_complete_cb(void *ctx);

int
owlap_init(wl_fw_read_cb_t *fw_read_cb, 
           void *fw_read_ctx,
           owlap_link_cb_t *link_cb, 
           void *link_ctx)
{
        enum wl_host_attention_mode mode;
        wl_err_t err;
        
        priv = owl_os_zalloc(sizeof(struct owlap_ctx));
        if (priv == NULL)
                return OWL_ERR_MEM;
        
        owl_dbg("fw download");
        err = wl_transport_init(fw_read_cb, fw_read_ctx, &mode);
        if (err != WL_SUCCESS) {
                owl_err("fw download err:%d", err);
                return -1;
        }

        owl_dbg("init");
        err = wl_init(NULL, init_complete_cb, mode);
        if (err != WL_SUCCESS) {
                owl_err("init err:%d", err);
                return -1;
        }
        
        owl_dbg("poll until init is complete");
        while (!priv->init_completed)
                wl_poll();

        wl_register_event_cb(event_cb, priv);
        priv->link_cb = link_cb;
        priv->link_ctx = link_ctx;

        owl_dbg("adding wifi tick timer");
        owltmr_addtmo(100, 1, wifi_tmr_cb, priv);

        owl_dbg("init done");
        return 0;
}

int
owlap_start(const struct wl_ssid_t *ssid, 
            const char *passphrase, 
            int channel,
            enum wl_enc_type enc_type,
            enum wl_auth_mode auth_mode,
            owlap_rand_cb_t rand_cb)
{
        wl_err_t wl_err;

        if (priv == NULL) {
                owl_info("ap is not initialized yet");
                return OWL_ERR_STATE;
        }
        
        owl_dbg("starting ap ssid:%s ssid_len:%d passphrase:%s channel:%d "
                "enc:%d auth:%d",
                ssid->ssid, ssid->len, passphrase ? passphrase : "-", channel,
                enc_type, auth_mode);

        if (ssid == NULL) {
                owl_err("ssid was NULL");
                return OWL_ERR_PARAM;
        }

        if (priv->running) {
                owl_info("ap already running, restarting ap");
        }
        
        switch (auth_mode) {
        case AUTH_MODE_SHARED_KEY:
        case AUTH_MODE_WPA:
        case AUTH_MODE_WPA2:
                owl_err("auth_mode:%d not supported", auth_mode);
                return OWL_ERR_NOTSUPP;
        default:
                break;
        }

        switch (enc_type) {
        case ENC_TYPE_WEP:
                owl_err("enc_type:%d not supported", enc_type);
                return OWL_ERR_NOTSUPP;
        default:
                break;
        }

        if (AUTH_MODE_OPEN_SYSTEM == auth_mode) {
                if ((wl_err = wlap_start_ap(ssid->ssid, 
                                            ssid->len, 
                                            channel)) < 0) {
                        owl_err("fail to start ap err:%d", wl_err);
                        return OWL_FAILURE;
                }
                
        } else {
#if defined(OWL_CONFIG_HOSTAPD)
                if (NULL == passphrase) {
                        owl_err("Passphrase required for WEP/WPA/WPA2");
                        return OWL_ERR_PARAM;
                }

                owl_dbg("starting hostapd");
                hostapd_start(ssid,
                              enc_type,
                              auth_mode,
                              channel,
                              passphrase,
                              rand_cb);
#else
                owl_err("hostapd not supported");
                return OWL_ERR_NOTSUPP;
#endif
        }
        priv->running = 1;
        owl_dbg("ap started");
        return 0;
}


int
owlap_stop(void)
{
        wl_err_t err;

        owl_dbg("stop");
        
        /* already stopped? nothing to do */
        if (!priv->running) {
                owl_info("ap already stopped, nothing to do");
                return 0;
        }
        
        if ((err = wlap_start_ap(NULL, 0, 1)) < 0) {
                owl_err("err:%d", err);
                return OWL_FAILURE;
        }

        priv->running = 0;
#if defined(OWL_CONFIG_HOSTAPD)
        hostapd_stop();
#endif

        return 0;
}


int
owlap_running(void)
{
        return priv->running;
}


static void
wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        (void) tmo; /* unused */
        (void) ctx; /* unused */
    
        wl_tick(owltmr_get_ms());
#if defined(OWL_CONFIG_HOSTAPD)
        /* Safe even if hostapd is inactive (not started) */
        hostapd_poll();
#endif
}


static void
event_cb(struct wl_event_t event, void *ctx)
{
        (void) ctx; /* unused */
        
        switch (event.id) {
        case WL_EVENT_MEDIA_CONNECTED:
                owl_dbg("first client connected");
                if (priv->link_cb)
                        priv->link_cb(priv->link_ctx, 1);
                break;
                
        case WL_EVENT_MEDIA_DISCONNECTED:
                owl_dbg("last client disconnected");
                if (priv->link_cb)
                        priv->link_cb(priv->link_ctx, 0);
                break;

        default:
                owl_err("unhandled event event.id:%d", event.id);
        };
}

static void
init_complete_cb(void *ctx)
{
        (void) ctx; /* unused */
        
        owl_dbg("ok");
        priv->init_completed = 1;
}


/**
 * connect shell command.
 *
 */
static void
ap_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct wl_ssid_t ssid;
        const char *passphrase = NULL;
        int channel = 1;
        enum wl_enc_type enc_type = ENC_TYPE_NONE;
        enum wl_enc_type auth_mode = AUTH_MODE_OPEN_SYSTEM;
        int len;
        int err;
        
        if (argc < 2)
                goto usage;

        if ((len = strlen(argv[1])) > WL_SSID_MAX_LENGTH)
                goto usage;
        
        memcpy(ssid.ssid, argv[1], len);
        ssid.len = len;

        /* channel */
        if (argc > 2)
                channel = atoi(argv[2]);

        if (argc > 3) {
                if (!owl_os_strcasecmp(argv[3], "none"))
                        enc_type = ENC_TYPE_NONE;
                else if (!owl_os_strcasecmp(argv[3], "tkip"))
                        enc_type = ENC_TYPE_TKIP;
                else if (!owl_os_strcasecmp(argv[3], "ccmp"))
                        enc_type = ENC_TYPE_CCMP;
                else
                        goto usage;
        }

        if (argc > 4) {
                if (!owl_os_strcasecmp(argv[4], "open"))
                        auth_mode = AUTH_MODE_OPEN_SYSTEM;
                else if (!owl_os_strcasecmp(argv[4], "wpa"))
                        auth_mode = AUTH_MODE_WPA_PSK;
                else if (!owl_os_strcasecmp(argv[4], "wpa2"))
                        auth_mode = AUTH_MODE_WPA2_PSK;
                else
                        goto usage;
        }
        
        if (argc > 5)
                passphrase = argv[5];

        if ((err = owlap_start(&ssid, passphrase, channel,
                               enc_type, auth_mode, NULL)))
                owlsh_printf(sh, "failed to start ap, err:%d", err);

        return;
        
usage:
        owlsh_printf(sh, "usage: ap <ssid> [channel] [enc] "
                     "[auth] [passphrase]\n\r");

}

const struct owlsh_cmd cmd_ap = {
        "ap", ap_f, "start ap"
};
