/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <lwip/init.h>
#include <lwip/stats.h>
#include <lwip/dhcp.h>
#include <lwip/dns.h>
#include <ipv4/lwip/ip_frag.h>
#include <ipv4/lwip/ip_addr.h>
#include <netif/etharp.h>
#include <netif/wlif.h>
#include <lwip/tcp_impl.h>

#include <owl/core/owl_net.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_os.h>

#if !NO_SYS
# include <lwip/tcpip.h>
#endif


/**
 * Setup and execution of lwip TCP/IP stack.
 *
 *
 *
 */


struct owlnet_ctx {
#define DHCP_STATE_IDLE      0
#define DHCP_STATE_RUN       1
        uint8_t dhcp_state;

#define FLAG_DHCP (1 << 0)
        uint8_t flags;

        void (*ipstatus)(struct netif *, void *ctx);
        void *ctx;

        struct netif *netif;
#if NO_SYS
        struct owltmr_tmo tcp_tmo;
        struct owltmr_tmo dhcp_tmo;
        uint32_t tick;
#endif
};

static struct owlnet_ctx *priv = NULL;

#if NO_SYS
static void
tcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        owl_unused(tmo);
        owl_unused(ctx);

        priv->tick++;

        tcp_tmr();
        if (!(priv->tick % (5000 / TCP_TMR_INTERVAL)))
                etharp_tmr();
        if (!(priv->tick % (DHCP_FINE_TIMER_MSECS / TCP_TMR_INTERVAL)))
                dhcp_fine_tmr();
        /* The dhcp_coarse_tmr() is called from dhcp_tmr_cb() since we need
           to call this also when link is down */
        if (!(priv->tick % (IP_TMR_INTERVAL / TCP_TMR_INTERVAL)))
                ip_reass_tmr();
#if LWIP_DNS == 1
        if (!(priv->tick % (DNS_TMR_INTERVAL / TCP_TMR_INTERVAL)))
                dns_tmr();
#endif
}

static void
dhcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        owl_unused(tmo);
        owl_unused(ctx);

        dhcp_coarse_tmr();
}


#if !OWL_CONFIG_WLX_API
static void
wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        owl_unused(tmo); /* unused */
        owl_unused(ctx); /* unused */

        wlif_poll(owlnet_netif());
}
#endif /* !OWL_CONFIG_WLX_API */
#endif /* NO_SYS */

static void
ipstatus_cb(struct netif *netif)
{
        owl_info("netif:%p up:%d", netif, netif_is_up(netif));

        if (priv->ipstatus)
                priv->ipstatus(netif, priv->ctx);
}

/* set static configuration and disable dhcp */
#if __GNUC__
# pragma GCC diagnostic ignored "-Wcast-qual"
#endif
void
owlnet_set_addr(const struct ip_addr *ip, const struct ip_addr *netmask,
                const struct ip_addr *gw, const struct ip_addr *dns)
{
        struct netif *netif;

        if (priv == NULL)
                return;

        netif = owlnet_netif();

        if (priv->dhcp_state != DHCP_STATE_IDLE) {
                owl_info("setting static ip, release dhcp");
                dhcp_release(netif);
                dhcp_stop(netif);
                priv->dhcp_state = DHCP_STATE_IDLE;
                owl_bitclr(priv->flags, FLAG_DHCP);
        }

        if (ip && !ip_addr_isany(ip))
                netif_set_ipaddr(netif, (struct ip_addr *) ip);
        if (netmask && !ip_addr_isany(netmask))
                netif_set_netmask(netif, (struct ip_addr *) netmask);
        if (gw && !ip_addr_isany(gw))
                netif_set_gw(netif, (struct ip_addr *) gw);
#if LWIP_DNS == 1
        if (dns && !ip_addr_isany(dns))
                dns_setserver(0, (struct ip_addr *) dns);
#endif

        if (netif_is_link_up(netif))
                netif_set_up(netif);
}
#if __GNUC__
# pragma GCC diagnostic warning "-Wcast-qual"
#endif

void
owlnet_linkup(void)
{
        struct netif *netif = owlnet_netif();
        netif_set_link_up(netif);

        if (owl_bit(priv->flags, FLAG_DHCP)) {

                if (priv->dhcp_state == DHCP_STATE_IDLE) {

                        struct ip_addr ipaddr, netmask, gw;
                        IP4_ADDR(&gw, 0,0,0,0);
                        IP4_ADDR(&ipaddr, 0,0,0,0);
                        IP4_ADDR(&netmask, 0,0,0,0);
                        netif_set_ipaddr(netif, &ipaddr);
                        netif_set_netmask(netif, &netmask);
                        netif_set_gw(netif, &gw);

                        owl_info("linkup, requesting dhcp");
                        dhcp_start(netif);
                        priv->dhcp_state = DHCP_STATE_RUN;
                }

        } else if (!ip_addr_isany(&netif->ip_addr)) {
                /* bring the interface up if a static ip is configured */

                owl_info("linkup, set interface up");
                netif_set_up(netif);

        } else {
                owl_info("linkup, netif still down");
        }

        owlnet_start(); /* compatibility with apps that does not call
                           owlnet_start() */
}

void
owlnet_linkdown(void)
{
        struct netif *netif = owlnet_netif();
        netif_set_link_down(netif);

#if OWL_CONFIG_NET_COMPAT
        /* To ensure that old applicaitions can still rely on ip_status
           changes when link goes down */
        owlnet_shutdown();
#endif
}

struct netif *
owlnet_init(void (*ipstatus)(struct netif *, void *ctx), void *ctx)
{
        struct ip_addr ipaddr, netmask, gw;

        if (priv) {
                owl_info("already initialized");
                wlif_reinit(owlnet_netif());
                return priv->netif;
        }

        IP4_ADDR(&gw, 0,0,0,0);
        IP4_ADDR(&ipaddr, 0,0,0,0);
        IP4_ADDR(&netmask, 0,0,0,0);

        /* With OS 'owl', this is done in owl_os_init() */
#if NO_SYS
# if !OWL_CONFIG_OS_OWL
        lwip_init();
# endif
#else
        tcpip_init(NULL, NULL);
#endif /* NO_SYS */

        priv = owl_os_zalloc(sizeof(*priv));
        if (priv == NULL)
                return NULL;

        priv->ipstatus = ipstatus;
        priv->ctx = ctx;

        priv->netif = owl_os_zalloc(sizeof(struct netif));
        if (priv->netif == NULL) {
                owl_os_free(priv);
                return NULL;
        }

        if ((netif_add(
		     priv->netif, &ipaddr, &netmask, &gw, NULL, wlif_init,
#if NO_SYS
		     ethernet_input
#else
		     tcpip_input
#endif /* NO_SYS */
		     )) == NULL) {
                owl_os_free(priv->netif);
                owl_os_free(priv);
                return NULL;
        }
        netif_set_default(priv->netif);
        netif_set_status_callback(priv->netif, ipstatus_cb);

#if NO_SYS
        owltmr_set_name(&priv->tcp_tmo, "tcp");
        owltmr_set_name(&priv->dhcp_tmo, "dhcp");
        owltmr_sched(
                &priv->dhcp_tmo, DHCP_COARSE_TIMER_MSECS, 1, dhcp_tmr_cb, NULL);
# if !OWL_CONFIG_WLX_API
        owltmr_addtmo(0, 1, wifi_tmr_cb, NULL);
# endif /* !OWL_CONFIG_WLX_API */
#endif /* NO_SYS */

        return priv->netif;
}

void
owlnet_shutdown(void)
{
        struct netif *netif = owlnet_netif();
        netif_set_down(netif);
}

void
owlnet_start(void)
{
#if NO_SYS
        owltmr_sched(&priv->tcp_tmo, TCP_TMR_INTERVAL, 1, tcp_tmr_cb, NULL);
#endif
}

void
owlnet_stop(void)
{
#if NO_SYS
        owltmr_cancel(&priv->tcp_tmo);
#endif
}


#if __GNUC__
# pragma GCC diagnostic ignored "-Wcast-qual"
#endif

struct netif *
owlnet_netif(void)
{
        return priv->netif;
}

#if __GNUC__
# pragma GCC diagnostic warning "-Wcast-qual"
#endif


/**
 *
 * Enable/disable dhcp. If we're connected, the dhcp address might be renewed or
 * released.
 *
 */
void
owlnet_set_dhcp(int enable)
{
        struct netif *netif;

        if (priv == NULL)
                return;

        netif = owlnet_netif();

        if (enable)
                owl_bitset(priv->flags, FLAG_DHCP);
        else
                owl_bitclr(priv->flags, FLAG_DHCP);

        if (netif == NULL)
                return;

        if (netif_is_link_up(netif)) {

                if (!owl_bit(priv->flags, FLAG_DHCP) &&
                    priv->dhcp_state != DHCP_STATE_IDLE) {
                        owl_info("link is up and dhcp disabled, relea dhcp");
                        dhcp_release(netif);
                        dhcp_stop(netif);
                        priv->dhcp_state = DHCP_STATE_IDLE;
                }

                else if (owl_bit(priv->flags, FLAG_DHCP) &&
                         priv->dhcp_state == DHCP_STATE_IDLE) {
                        owl_info("link is up and dhcp enabled, request dhcp");
                        netif_set_down(netif); /* static ip no longer valid */
                        dhcp_start(netif);
                        priv->dhcp_state = DHCP_STATE_RUN;
                }
        }
}


int
owlnet_get_dhcp(void)
{
        return owl_bit(priv->flags, FLAG_DHCP);
}

/**
 * Hostnames are composed of series of labels concatenated with dots. Each
 * label must be between 1 and 63 characters long, and the entire hostname
 * (including the delimiting dots) has a maximum of 255 characters.
 *
 * Component hostname labels may contain only the ASCII letters 'a' through 'z'
 * (in a case-insensitive manner), the digits '0' through '9', and the hyphen
 * ('-').
 *
 */
int
owlnet_is_hostname(const char *str)
{
        const char *iter;
        int len = 0;
        int label = 0;

        if (str == NULL)
                return 0;

        for (iter = str; *iter; iter++) {
                if (owl_islower(*iter) ||
                    owl_isupper(*iter) ||
                    owl_isdigit(*iter) ||
                    *iter == '-') {
                        label++;

                } else if (*iter == '.') {
                        if (label == 0)
                                return 0;

                        label = 0;

                } else {

                        return 0;
                }

                if (label > 63)
                        return 0;

                len++;
                if (len > 255)
                        return 0;
        }

        if (len == 0)
                return 0;

        /* can't end with '.' */
        if (label == 0)
                return 0;

        return 1;
}


static void
ifconfig_f(struct owlsh_ctx *ctx, int argc, const char *argv[])
{
        struct netif *netif = owlnet_netif();
        struct ip_addr addr;

        /* check args */
        if (argc > 1 && !ip_aton(argv[1], &addr))
                goto usage;

        if (argc > 2 && !ip_aton(argv[2], &addr))
                goto usage;

        if (argc > 3 && !ip_aton(argv[3], &addr))
                goto usage;


        /* do something */
        if (argc > 1 && ip_aton(argv[1], &addr))
                netif_set_ipaddr(netif, &addr);

        if (argc > 2 && ip_aton(argv[2], &addr))
                netif_set_netmask(netif, &addr);

        if (argc > 3 && ip_aton(argv[3], &addr))
                netif_set_gw(netif, &addr);

#if LWIP_DNS
        if (argc > 4 && ip_aton(argv[4], &addr))
                dns_setserver(0, &addr);
#endif

        if (argc > 1)
                return;

        owlsh_printf(ctx, "%c%c    HWaddr %02x:%02x:%02x:%02x:%02x:%02x\n\r",
                   netif->name[0], netif->name[1],
                   netif->hwaddr[0], netif->hwaddr[1], netif->hwaddr[2],
                   netif->hwaddr[3], netif->hwaddr[4], netif->hwaddr[5]);

        owlsh_printf(ctx, "      inet addr:%d.%d.%d.%d  mask:%d.%d.%d.%d "
                     "gw:%d.%d.%d.%d\n\r",
                     ip4_addr1(&netif->ip_addr), ip4_addr2(&netif->ip_addr),
                     ip4_addr3(&netif->ip_addr), ip4_addr4(&netif->ip_addr),
                     ip4_addr1(&netif->netmask), ip4_addr2(&netif->netmask),
                     ip4_addr3(&netif->netmask), ip4_addr4(&netif->netmask),
                     ip4_addr1(&netif->gw), ip4_addr2(&netif->gw),
                     ip4_addr3(&netif->gw), ip4_addr4(&netif->gw));

#if LWIP_DNS
        addr = dns_getserver(0);
        owlsh_printf(ctx, "           dns:%d.%d.%d.%d\n\r",
                     ip4_addr1(&addr), ip4_addr2(&addr),
                     ip4_addr3(&addr), ip4_addr4(&addr));
#endif

        owlsh_printf(ctx, "      %s%s%s%s%s%s%sMTU:%d\n\r",
                   netif->flags & NETIF_FLAG_UP ? "UP " : "",
                   netif->flags & NETIF_FLAG_BROADCAST ? "BROADCAST " : "",
                   netif->flags & NETIF_FLAG_POINTTOPOINT ? "PPP " : "",
                   netif->flags & NETIF_FLAG_DHCP ? "DHCP " : "",
                   netif->flags & NETIF_FLAG_LINK_UP ? "LINK-UP " : "",
                   netif->flags & NETIF_FLAG_ETHARP ? "ETHARP " : "",
                   netif->flags & NETIF_FLAG_IGMP ? "IGMP " : "",
                   netif->mtu);
#if LWIP_STATS && LINK_STATS
        owlsh_printf(ctx, "      RX packets:%d\n\r", lwip_stats.link.recv);
        owlsh_printf(ctx, "      TX packets:%d\n\r", lwip_stats.link.xmit);
#endif /* LWIP_STATS && LINK_STATS */
        return;

usage:
        owlsh_printf(ctx, "usage: ifconfig [ipaddr] [mask] [gw]\n\r");

}

static void
dhcp_f(struct owlsh_ctx *ctx, int argc, const char *argv[])
{
        if (argc > 1) {
                if (owl_os_strcmp("on", argv[1]) == 0)
                        owlnet_set_dhcp(1);
                else if (owl_os_strcmp("off", argv[1]) == 0)
                        owlnet_set_dhcp(0);
                else
                        goto usage;

                return;
        }

        owlsh_printf(ctx, "dhcp:%s\n\r", owlnet_get_dhcp() ? "on" : "off");
        return;

usage:
        owlsh_printf(ctx, "usage: dhcp [on|off]\n\r");
}

#if 0 /* need for < lwip-1.4.0 */
int
ip_aton(const char *cp, struct ip_addr *addr)
{
        struct in_addr ip;
        if (!inet_aton(cp, &ip))
                return 0;

        addr->addr = ip.s_addr;
        return 1;
}
#endif

int
lwip_error(err_t err)
{
        switch (err) {
        case ERR_OK:
                return 0;
        case ERR_ISCONN: /* fall through */
        case ERR_RTE:
                return OWL_ERR_STATE;
        case ERR_VAL:
                return OWL_ERR_PARAM;
        case ERR_MEM:
                return OWL_ERR_MEM;
        case ERR_USE:
                return OWL_ERR_INUSE;
        case ERR_INPROGRESS:
                return OWL_ERR_BUSY;
        default:
                owl_err("unknown lwip return code:%d", err);
                return OWL_FAILURE;
        };
}


const struct owlsh_cmd cmd_ifconfig = {
        "ifconfig", ifconfig_f, "config ip address"
};

const struct owlsh_cmd cmd_dhcp = {
        "dhcp", dhcp_f, "enable dhcp"
};
