/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

#include <wl_api.h>

#include <owl/core/owl_wifi.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>

#include <wl_os.h>

struct owlwifi_ctx {
        struct wl_ssid_t ssid;
        struct wl_mac_addr_t bssid;

        owlwifi_link_cb_t *link_cb;
        void *link_ctx;

        int running;
        enum owlwifi_connect_state state;
        volatile int init_completed;
};

static struct owlwifi_ctx *priv = NULL;

static void wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx);
static int is_ssid_set(void);
static int is_bssid_set(void);
static int is_ssid_equal(struct wl_ssid_t *s1, struct wl_ssid_t *s2);
static int is_bssid_equal(struct wl_mac_addr_t *b1, struct wl_mac_addr_t *b2);
static void connect(void);
static void event_cb(struct wl_event_t event, void *ctx);
static void init_complete_cb(void *ctx);
static int ascii_to_key(char *outp, const char *inp);

/**
 * Initialize wifi device and connection manager.
 *
 */
int
owlwifi_init(wl_fw_read_cb_t *fw_read_cb, void *fw_read_ctx,
             owlwifi_link_cb_t *link_cb, void *link_ctx)
{
        enum wl_host_attention_mode mode;
        wl_err_t err;
        
        priv = owl_os_zalloc(sizeof(struct owlwifi_ctx));
        if (priv == NULL)
                return -1;
        
        owl_dbg("fw download");
        err = wl_transport_init(fw_read_cb, fw_read_ctx, &mode);
        switch (err) {
        case WL_SUCCESS:
                break;
                
        case WL_CARD_FAILURE:
                owl_err("could not detect wifi device, err:%d", err);
                return -1;
                
        case WL_FIRMWARE_INVALID:
                owl_err("invalid wifi firmware image, err:%d", err);
                return -1;
                
        default:
                return -1;
        };
        

        owl_dbg("init");
        if ((err = wl_init(NULL, init_complete_cb, mode)) != WL_SUCCESS) {
                owl_err("wifi initialization failed with err:%d", err);
                return -1;
        }

        owl_dbg("poll until init is complete");
        while (!priv->init_completed)
                wl_poll();

        wl_register_event_cb(event_cb, priv);
        priv->link_cb = link_cb;
        priv->link_ctx = link_ctx;

        owl_dbg("adding wifi tick timer");
        owltmr_addtmo(100, 1, wifi_tmr_cb, priv);

        memset(&priv->bssid, 0xff, sizeof(priv->bssid));
        wl_scan(NULL);
        owl_dbg("init done");
        return 0;
}

const char *
owlwifi_ssid_str(const struct wl_ssid_t *ssid)
{
        static char ssid_str[WL_SSID_MAX_LENGTH];

        if (ssid == NULL) {
                *ssid_str = 0;
                return ssid_str;
        }
        
        owl_os_memcpy(ssid_str, ssid->ssid, ssid->len);
        ssid_str[ssid->len] = 0;
        return ssid_str;
}


static int set_wep_key(const char *key,
                       struct wl_network_t *net) 
{
        int keylen, err, i;
        char wepkey[13];

        keylen = strlen(key);
        
        if (keylen == 5 || keylen == 13) {
                /* Passphrase? */
                strncpy(wepkey, key, keylen);
                
        } else {
                /* Otherwise it's a hex string */
                keylen = ascii_to_key(wepkey, key);
                if (keylen != 5 && keylen != 13) {
                        owl_err("invalid wep key len: %d", keylen);
                        return OWL_ERR_PARAM;
                }
        }
        
        for (i = 0; i < 4; i++) {
                if ((err = wl_delete_wep_key(i, &net->bssid)) !=
                    WL_SUCCESS) {
                        owl_err("fail to del wep key %d, err:%d", i,
                                err);
                        return OWL_FAILURE;
                }
        }
        
        owl_dbg("adding wep key of length %d\r", keylen);
        if ((err = wl_add_wep_key(0, keylen, wepkey,
                                  &net->bssid)) != WL_SUCCESS) {
                owl_err("failed to add wep key: %s err: %d",
                        key, err);
                return OWL_FAILURE;
        }
        
        if ((err = wl_set_default_wep_key(0)) != WL_SUCCESS) {
                owl_err("failed to set default wep key, err:%d",
                        err);
                return OWL_FAILURE;
        }
        return 0;
}

/**
 * Set wifi configuration. This is typically called from a db callback.
 * All parameters will always be set.
 * 
 */
int
owlwifi_set_conf(const struct wl_ssid_t *ssid,
                 const struct wl_mac_addr_t *bssid,
                 const char *key, int auth)
{
        struct wl_mac_addr_t bcast_bssid;
        wl_err_t err;
        struct wl_network_t net;
        int keylen, auth_mode;
        
        
#if OWL_TRACE >= OWL_TRACE_INFO
        if (bssid)
                owl_info("ssid:%s bssid:%02x%02x%02x%02x%02x%02x key:%s auth:%d",
                         owlwifi_ssid_str(ssid),
                         bssid->octet[0], bssid->octet[1],
                         bssid->octet[2], bssid->octet[3],
                         bssid->octet[4], bssid->octet[5],
                         key ? key : "NULL", auth);
        else
                owl_info("ssid:%s bssid:NULL key:%s auth:%d",
                         owlwifi_ssid_str(ssid), key ? key : "NULL", auth);
            
#endif
        
        if (priv == NULL)
                return OWL_ERR_STATE;

        memset(&bcast_bssid, 0xff, sizeof(bcast_bssid));
        
        if (ssid) {
                memset(&priv->ssid, 0, sizeof(priv->ssid));
                memcpy(&priv->ssid, ssid, sizeof(priv->ssid));
        }
        
        if (bssid) {
                memset(&priv->bssid, 0xff, sizeof(priv->bssid));
                memcpy(&priv->bssid, bssid, sizeof(priv->bssid));
        }

        if (key == NULL) {
                connect();
                return 0;
        }

        keylen = strlen(key);
        
        /* can not have keys of zero length */
        if (keylen == 0)
                return OWL_ERR_PARAM;
        
        /* cannot set wpa key for a specific bssid */
        if (bssid && auth > OWLWIFI_WEP)
                return OWL_ERR_NOTSUPP;
                
        memset(&net, 0, sizeof net);
        memset(&net.bssid, 0xff, sizeof(net.bssid));
        net.enc_type = ENC_TYPE_AUTO;
        memcpy(&net.ssid, &priv->ssid, sizeof(net.ssid));
        
        auth_mode = AUTH_MODE_INVALID;
        switch (auth) {
        case OWLWIFI_WEP:
                err = set_wep_key(key, &net);
                if (err < 0)
                        return err;
                break;
        case OWLWIFI_AUTH_AUTO:
                auth_mode = AUTH_MODE_AUTO;
                break;
        case OWLWIFI_WPA_PSK:
                auth_mode = AUTH_MODE_WPA_PSK;
                break;
        case OWLWIFI_WPA2_PSK:
                auth_mode = AUTH_MODE_WPA2_PSK;
                break;

        default :
                owl_err("Bad auth %d", auth);
                return OWL_FAILURE;
        }
        if (auth_mode > AUTH_MODE_INVALID) {
                owl_dbg("set passphrase of len %d", keylen);

                if ((err = wl_set_passphrase(&net, key, keylen, ENC_TYPE_AUTO,
                                             auth_mode)) != WL_SUCCESS) {
                        owl_err("failed to add passhrase: %s, err:%d",
                                key, err);
                        return OWL_FAILURE;
                }
        }

        connect();
        return 0;
}


/**
 * Start the connection manager.
 * 
 * The ssid and bssid of the desired network should be specified. The ssid and
 * bssid will be matched against the networks found during scan. bssid will
 * have priority over ssid.
 *
 * @param ssid The ssid of the desired network.
 * @param bssid The bssid of the desired network.
 *          
 */
int
owlwifi_start(void)
{
        owl_info("enter");

        priv->running = 1;

        /* start scan, do things in the event_cb() when scan is completed */
        wl_scan(NULL);
        return 0;
}

int
owlwifi_stop(void)
{
        owl_info("enter");

        priv->running = 0;
        connect();
        return 0;
}

int
owlwifi_running(void)
{
        return priv->running;
}


enum owlwifi_connect_state 
owlwifi_conn_state(void) 
{
        return priv->state;
}


static void
wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
    (void) tmo; /* unused */
    (void) ctx; /* unused */
    
    wl_tick(owltmr_get_ms());
}


static void
event_cb(struct wl_event_t event, void *ctx)
{
        (void) ctx; /* unused */
        
        switch (event.id) {
        case WL_EVENT_MEDIA_CONNECTED:
                owl_dbg("media connected");
                priv->state = OWLWIFI_CONNECTED;
                if (priv->link_cb)
                        priv->link_cb(priv->link_ctx, 1);
                break;
                
        case WL_EVENT_CONN_FAILURE:
                priv->state = OWLWIFI_DISCONNECTED;
                owl_dbg("conn failure");
                wl_scan(NULL);
                break;
                
        case WL_EVENT_MEDIA_DISCONNECTED:
                priv->state = OWLWIFI_DISCONNECTED;
                owl_dbg("media disconnected");
                if (priv->link_cb)
                        priv->link_cb(priv->link_ctx, 0);
                wl_scan(NULL);
                break;

        case WL_EVENT_SCAN_COMPLETE:
                owl_dbg("scan complete");
                connect();
                break;

        default:
                owl_err("unhandled event");
        };
}


/**
 * Connection manager decisions are taken here.
 *
 */
static void
connect(void)
{
        struct wl_network_t *net;
        wl_err_t err;
        
        net = wl_get_current_network();

        if (!priv->running) {
                owl_dbg("not running");
                
                if (net) {
                        owl_dbg("disconnect to stop");
                        err = wl_disconnect();
                        if (err != WL_SUCCESS && err != WL_NOEFFECT)
                                goto fail;
                }

        } else if (net == NULL && is_bssid_set()) {
                
                owl_dbg("not connected and bssid is set, try to connect");
                
                /* Connect to bssid */
                if ((err = wl_connect_bssid(priv->bssid)) != WL_SUCCESS)
                        goto fail;
                priv->state = OWLWIFI_CONNECTING;

        } else if (net == NULL && is_ssid_set()) {

                owl_dbg("not connected and ssid is set, try to connect");
                
                /* Connect to ssid */
                if ((err = wl_connect(priv->ssid.ssid,
                                      priv->ssid.len)) != WL_SUCCESS)
                        goto fail;
                priv->state = OWLWIFI_CONNECTING;
                
        } else if (net && !is_ssid_set() && !is_bssid_set()) {

                owl_dbg("connected and neither ssid or bssid is set, "  \
                        "disconnect");
                
		/* Disconnect if we're connected and ssid and bssid
		 * is not configured.
		 */
                if ((err = wl_disconnect()) != WL_SUCCESS)
                        goto fail;

        } else 	if (net &&
		    ((is_ssid_set() &&
                      !is_ssid_equal(&net->ssid, &priv->ssid)) ||
		     (is_bssid_set() &&
                      !is_bssid_equal(&net->bssid, &priv->bssid)))) {

                owl_dbg("connected and ssid or bssid has changed, "     \
                        "disconnect");

                /* Change network */
                if ((err = wl_disconnect()) != WL_SUCCESS)
                        goto fail;
        } else {
                owl_dbg("nothing to do");
        }

        return;

fail:
        owl_err("err %d", err);
        wl_scan(NULL);
}        


static void
init_complete_cb(void *ctx)
{
        (void) ctx; /* unused */
        
        owl_dbg("ok");
        priv->init_completed = 1;
}

static int
is_ssid_equal(struct wl_ssid_t *s1, struct wl_ssid_t *s2)
{
        if (s1->len != s2->len)
                return 0;

        if (memcmp(s1, s2, s1->len))
                return 0;

        return 1;
}

static int
is_bssid_equal(struct wl_mac_addr_t *b1, struct wl_mac_addr_t *b2)
{
        if (memcmp(b1, b2, sizeof(*b1)))
                return 0;

        return 1;
}

static int
is_ssid_set(void)
{
        return priv->ssid.len > 0;
}

static int
is_bssid_set(void)
{
        return memcmp(&priv->bssid, "\xff\xff\xff\xff\xff\xff",
                      sizeof(priv->bssid)) != 0;
}


/**
 * Helper function to print information about a wifi network.
 *
 * @param net is the network for which information will be shown.
 */
static void
print_network(struct owlsh_ctx *sh, struct wl_network_t *net)
{
	char buf[WL_SSID_MAX_LENGTH + 1];

	/* bssid */
	owlsh_printf(sh, "%02x:%02x:%02x:%02x:%02x:%02x ",
                     net->bssid.octet[0], net->bssid.octet[1],
                     net->bssid.octet[2], net->bssid.octet[3],
                     net->bssid.octet[4], net->bssid.octet[5]);

	/* ssid */
	memset(buf, 0, sizeof buf);
	memcpy(buf, net->ssid.ssid, net->ssid.len);
	owlsh_printf(sh, "\"%s\" ", buf);

	/* rssi */
	owlsh_printf(sh, "RSSI %d dBm ", (int) net->rssi);

        /* auth */
        switch (net->auth_mode) {
        case AUTH_MODE_WPA:
		owlsh_printf(sh, "(WPA) ");
                break;
        case AUTH_MODE_WPA2:
		owlsh_printf(sh, "(WPA2) ");
                break;
        case AUTH_MODE_WPA_PSK:
		owlsh_printf(sh, "(WPA_PSK) ");
                break;
        case AUTH_MODE_WPA2_PSK:
		owlsh_printf(sh, "(WPA2_PSK) ");
                break;
        }
	/* enc */
	switch (net->enc_type) {
	case ENC_TYPE_WEP :
		owlsh_printf(sh, "(WEP)");
		break;
	case ENC_TYPE_TKIP :
		owlsh_printf(sh, "(TKIP)");
		break;
	case ENC_TYPE_CCMP :
		owlsh_printf(sh, "(CCMP)");
		break;
	case ENC_TYPE_NONE :
		break;
	}
	owlsh_printf(sh, "\n\r");
}


/**
 * scan shell command callback.
 * Note that the scan results presented will be from the last scan,
 * not this one.
 *
 */
static void
scan_f(struct owlsh_ctx *sh, int argc,  const char *argv[])
{
        unsigned int i;
        struct wl_network_list_t *netlist = NULL;

        (void) argc; /* unused */
        (void) argv; /* unused */
        
        wl_scan(NULL);
        if (wl_get_network_list(&netlist) != WL_SUCCESS) {
		owlsh_printf(sh, "failed to get network list\n\r");
		return;
        }

        if (netlist->cnt == 0)
                owlsh_printf(sh, "no nets found\n\r");

        for (i = 0; i < netlist->cnt; i++)
		print_network(sh, netlist->net[i]);
}


/**
 * connect shell command.
 *
 */
static void
connect_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct wl_ssid_t ssid;
        int len;
        
        if (argc < 2)
                goto usage;

        if ((len = strlen(argv[1])) > WL_SSID_MAX_LENGTH)
                goto usage;
        
        memcpy(ssid.ssid, argv[1], len);
        ssid.len = len;

        owlwifi_set_conf(&ssid, NULL, NULL, 0);
        owlwifi_start();
        return;

usage:
        owlsh_printf(sh, "usage: connect <ssid>\n\r");

}

/**
 * wpass shell command
 * Set a WPA passphrase
 *
 */
static void
setpass_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct wl_network_t net;
        const char *passphrase, *auth;
        int len, auth_type;

        if (argc < 3)
                goto usage;
        if (4 == argc) {
                auth = argv[3];
                if (strcmp(auth, "wpa2") == 0)
                        auth_type = AUTH_MODE_WPA2_PSK;
                else if (strcmp(auth, "wpa") == 0)
                        auth_type = AUTH_MODE_WPA_PSK;
                else
                        goto usage;
        }
        else 
                auth_type = AUTH_MODE_AUTO;
                
        if ((len = strlen(argv[1])) > WL_SSID_MAX_LENGTH)
                goto usage;

        memset(&net, 0, sizeof net);
        memset(net.bssid.octet, 0xff, sizeof net.bssid.octet);

        memcpy(net.ssid.ssid, argv[1], len);
        net.ssid.len = len;

        net.enc_type = ENC_TYPE_AUTO;
        passphrase = argv[2];
        
        if (wl_set_passphrase(&net, passphrase, strlen(passphrase), 
                              ENC_TYPE_AUTO, auth_type) != WL_SUCCESS)
                owlsh_printf(sh, "failed to add passphrase\n\r");

        return;
        
usage:
        owlsh_printf(sh, "usage: wpass <ssid> <passphrase> [wpa|wpa2]\n\r");
}


/**
 *
 */
static void
power_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        if (argc < 2)
                goto usage;
        
        if (!strcmp(argv[1], "on")) {
                if (wl_enable_ps() != WL_SUCCESS)
                        owlsh_printf(sh, "could not enable power save\n\r");

        } else if(!strcmp(argv[1], "off")) {
                if (wl_disable_ps() != WL_SUCCESS)
                        owlsh_printf(sh, "could not disable power save\n\r");
                
        } else {
                goto usage;
        }
                
        return;

usage:
        owlsh_printf(sh, "usage: powersave <on|off>\n\r");
}


/**
 *
 */
static void
psconf_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        uint8_t use_ps_poll;
        uint32_t traffic_timeout;
        uint32_t ps_delay;
        uint8_t rx_all_dtim;
        uint16_t listen_interval;
        
        if (argc < 6)
                goto usage;
        
        use_ps_poll = atoi(argv[1]);
        traffic_timeout = atoi(argv[2]);
        ps_delay = atoi(argv[3]);
        rx_all_dtim = atoi(argv[4]);
        listen_interval = atoi(argv[5]);

        if (use_ps_poll > 1)
                goto usage;

        if (rx_all_dtim > 1)
                goto usage;

        if (wl_conf_ps(use_ps_poll, traffic_timeout, ps_delay,
                       rx_all_dtim, listen_interval) != WL_SUCCESS)
                owlsh_printf(sh, "power save configuration failed\n\r");
        
        return;
        
usage:
        owlsh_printf(sh, "usage: psconf <use_ps_poll>  ([bool]   def 0)\n\r");
        owlsh_printf(sh, "              <traffic_tmo>  ([ms]     def 10)\n\r");
        owlsh_printf(sh, "              <ps_delay>     ([ms]     def 5000)\n\r");
        owlsh_printf(sh, "              <rx_all_dtim>  ([bool]   def 1)\n\r");
        owlsh_printf(sh, "              <listen_intvl> ([beacon] def 20)\n\r");
}


static int
ascii_to_key(char *outp, const char *inp) {
        char buf[3];
        int len;
        buf[2] = '\0';
        len = strlen(inp);
        if (len % 2)
                return 0;

        len = 0;
        while (*inp) {
                if (!isxdigit((unsigned char) *inp) ||
                    !isxdigit((unsigned char) *(inp+1)) ||
                    len > WL_MAX_PASS_LEN) {
                        return 0;
                }
                buf[0] = *inp++;
                buf[1] = *inp++;
                *outp++ = strtol(buf, NULL, 16);
                len++;
        }
        return len;
}


/**
 * setkey shell command.
 * Configure wep keys
 */
static void
setkey_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int idx, len;
        char key[13];
        struct wl_mac_addr_t bssid;

        memset(&bssid.octet, 0xff, sizeof bssid.octet);
        
        if (argc == 2 && strcmp(argv[1], "none") == 0) {
                owlsh_printf(sh, "Deleting WEP keys\n\r");
                wl_delete_wep_key(0, &bssid);
                wl_delete_wep_key(1, &bssid);
                wl_delete_wep_key(2, &bssid);
                wl_delete_wep_key(3, &bssid);
                return;
        }
        
        if (argc < 3)
                goto usage;

        idx = atoi(argv[1]);
        len = strlen(argv[2]);

        /* Pass phrase? */
        if (len == 5 || len == 13) {
                strncpy(key, argv[2], len);

        /* Otherwise it's a hex string */
        } else {

                len = ascii_to_key(key, argv[2]);
                if (len == 0 || idx > 3 || idx < 0 ||
                    (idx == 0 && *argv[1] != '0'))
                        goto usage;
                
                if (len != 5 && len != 13) {
                        owlsh_printf(sh, "WEP key must be 10 (WEP-40) or " \
                                     "26 (WEP-104) digits\n\r");
                        goto usage;
                }
        }
        
        wl_add_wep_key(idx, len, key, &bssid);
        wl_set_default_wep_key(idx);
        return;

usage:
        owlsh_printf(sh, "usage: setkey <key_idx (0-3)> <key in hex>\n\r\t " \
                     "or: setkey none\n\r");
}


const struct owlsh_cmd cmd_scan = {
        "scan", scan_f, "scan for networks"
};

const struct owlsh_cmd cmd_connect = {
        "connect", connect_f, "connect to network"
};

const struct owlsh_cmd cmd_wpass = {
        "wpass", setpass_f, "set wpa passphrase"
};

const struct owlsh_cmd cmd_powersave = {
        "powersave", power_f, "set powersave mode"
};

const struct owlsh_cmd cmd_psconf = {
        "psconf", psconf_f, "config powersave mode"
};

const struct owlsh_cmd cmd_setkey = {
        "setkey", setkey_f, "set wep key(s)"
};
