/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <owl/core/owl_list.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_net.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>

#include <owl/util/httpd/httpd.h>

#include <owl/util/httpd/httpd_404.h>
#ifdef OWL_CONFIG_HTTPD_PARSER
# include <owl/util/httpd/httpd_parser.h>
#endif
#ifdef OWL_CONFIG_HTTPD_AUTH
# include <owl/util/httpd/httpd_auth.h>
#endif

#include <lwip/tcp.h>

struct httpd_document {
        httpd_cb_t *f;
        void *ctx;

        enum http_method method;
        const char *path;
        
#ifdef OWL_CONFIG_HTTPD_AUTH
        struct http_auth *auth;
#endif
        
        struct list_head list;        
};

struct httpd_conn {
        httpd_cb_t *f;
        void *ctx;
        
        struct pbuf *pbuf; /* HTTP request */
        int seq;
        char *rspbuf;
        uint16_t rsplen;
        uint16_t offset; /* read offset */
        int left;
        uint8_t retries;
    
        struct httpd_ctx *httpd;
        void *local; /* connection-local data passed to documents */
};

struct httpd_ctx {
        uint16_t port;
        struct tcp_pcb *pcb;
        int run;
        int alignment;   /* tx data size alignment, must be power of 2 */

#ifdef OWL_CONFIG_HTTPD_AUTH
        struct http_auth auth;
#endif
        
        struct list_head documents;
};

static void conn_timer_cb(struct owltmr_tmo *tmo, void *ctx);

struct httpd_ctx *
httpd_create(uint16_t port)
{
        struct httpd_ctx *priv;
        
        if (port == 0)
                return NULL;
        
        priv = owl_os_zalloc(sizeof(struct httpd_ctx));
        if (priv == NULL)
                return NULL;

        if (owltmr_addtmo(100, 1, conn_timer_cb, priv) == NULL) {
                owl_os_free(priv);
                return NULL;
        }

        priv->alignment = 4;
        priv->port = port;
        list_init(&priv->documents);

        owl_dbg("created server at port:%d", port);
        
        return priv;
}

struct httpd_document *
httpd_add_document(struct httpd_ctx *priv,
                   enum http_method method, const char *path,
                   httpd_cb_t *cb, void *ctx)
{
        struct httpd_document *doc;

        if ((doc = owl_os_zalloc(sizeof(*doc))) == NULL)
                return NULL;

        doc->method = method;
        doc->path = path;
        doc->f = cb;
        doc->ctx = ctx;
        
        owl_info("adding document:0x%x", doc);
        
        list_add_tail(&doc->list, &priv->documents);
        return doc;
}

void
httpd_set_auth(struct httpd_ctx *priv,
                  enum http_auth_scheme scheme,
                  const char *user, const char *password,
                  const char *realm)
{
#ifdef OWL_CONFIG_HTTPD_AUTH
        priv->auth.scheme = scheme;
        priv->auth.user = user;
        priv->auth.password = password;
        priv->auth.realm = realm;
#else
        owl_unused(priv);
        owl_unused(scheme);
        owl_unused(user);
        owl_unused(password);
        owl_unused(realm);
#endif  /* OWL_CONFIG_HTTPD_AUTH */
}

int
httpd_set_document_auth(struct httpd_ctx *priv,
                           struct httpd_document *doc,
                           int scheme, const char *user, const char *password,
                           const char *realm)
{
#ifdef OWL_CONFIG_HTTPD_AUTH
        struct http_auth *auth;

        if ((auth = owl_os_zalloc(sizeof(*auth))) == NULL)
                return OWL_ERR_MEM;

        auth->scheme = scheme;
        auth->user = user;
        auth->password = password;
        auth->realm = realm;

        owl_os_free(doc->auth);
        doc->auth = auth;
        return 0;
#else
        owl_unused(priv);
        owl_unused(doc);
        owl_unused(scheme);
        owl_unused(user);
        owl_unused(password);
        owl_unused(realm);

        return 0;
#endif /* OWL_CONFIG_HTTPD_AUTH */                           
}


void
httpd_start(struct httpd_ctx *priv)
{
        owl_info("starting httpd");
        priv->run = 1;
}

void
httpd_stop(struct httpd_ctx *priv)
{
        owl_info("stopping httpd");
        priv->run = 0;
}

static void
conn_err(void *arg, err_t err)
{
        struct httpd_conn *conn = arg;
        
        owl_info("conn:0x%x", conn);
        owl_unused(err);
        
        if (conn) {
                owl_os_free(conn->rspbuf);
                owl_os_free(conn);
        }
}

static void
close_conn(struct tcp_pcb *pcb, struct httpd_conn *conn)
{
        owl_dbg("close conn:%p", conn);
        
        tcp_arg(pcb, NULL);
        tcp_sent(pcb, NULL);
        tcp_recv(pcb, NULL);
        if (conn->pbuf)
                pbuf_free(conn->pbuf);
        
        owl_os_free(conn->rspbuf);
        owl_os_free(conn);
        tcp_close(pcb);
}

static void
send_data(struct tcp_pcb *pcb, struct httpd_conn *conn, int more)
{
        err_t err;
        uint16_t len;
        int align = conn->httpd->alignment;
        
        owl_dbg("left:%d tcp_sndbuf:%d", conn->left, tcp_sndbuf(pcb));
        
        /* We cannot send more data than space available in the send buffer. */
        if (tcp_sndbuf(pcb) < conn->left)
                len = tcp_sndbuf(pcb);
        else
                len = conn->left;

        /* if alignment is set in the http configuration, the http server
           will ensure that all transmitted segments are size aligned. Any
           unaligned sized segment will appear as the last chunk. This is
           to support platforms which are not able to transmit data that is
           not size aligned. Any padding performed by wl_process_tx() only
           affects the complete IP packet - individual segments in a pbuf
           chain will not be padded. This feature can be disabled by setting
           httpd->alignment to 0 */
        if (align && more)
                len = len & ~(align - 1);
        
        do {
                err = tcp_write(pcb, conn->rspbuf + conn->offset, len,
                                TCP_WRITE_FLAG_COPY);
                
                if (err == ERR_MEM)
                        len /= 2;

        } while (err == ERR_MEM && len > 1);

  
        if (err == ERR_OK) {
                owl_dbg("wrote %d bytes", len);
                owl_hexdump("HTTP TX", conn->rspbuf + conn->offset, len);
                conn->offset += len;
                conn->left -= len;
                
        } else {
                owl_err("fail to write:%d", err);
                
        }
}

static err_t
http_sent(void *arg, struct tcp_pcb *pcb, u16_t len)
{
        struct httpd_conn *conn = arg;
        int align = conn->httpd->alignment;

        owl_unused(len);
        
        conn->retries = 0;
  
        if (conn->left > align) {
                send_data(pcb, conn, 1);
                
        } else {
                int res = 0;
                
                /* if we have any remaining bytes (due to size alignment), move
                   it to the front of the buffer so we can continue filling the
                   rest */
                if (conn->left) {
                        owl_info("%d bytes left in rspbuf, moving to front",
                                 conn->left);
                        owl_os_memcpy(conn->rspbuf, conn->rspbuf + conn->offset,
                                      conn->left);
                }
                
                /* Request more data from the file */
                if (conn->f) {
                        owl_info("requesting more data seq:%d", conn->seq);
                        res = conn->f(conn->ctx, conn->seq, NULL, NULL,
                                      conn->rspbuf + conn->left,
                                      conn->rsplen - conn->left, &conn->local);
                        conn->left += res;
                        conn->seq++;
                        if (0 == res)
                                conn->f = NULL;
                }
                if (conn->left > 0) {
                        owl_info("got %d bytes from handler:0x%x conn->left:%d",
                                 res, conn->f, conn->left);
                        conn->offset = 0;
                        send_data(pcb, conn, res);
                } else {
                        owl_info("got %d from handler:0x%x, closing conn",
                                 conn->left, conn->f);
                        close_conn(pcb, conn);
                }
        }

        return ERR_OK;
}

static err_t
http_poll(void *arg, struct tcp_pcb *pcb)
{
        struct httpd_conn *conn = arg;

        
        if (conn == NULL) {
                owl_dbg("aborting pcb");
                tcp_abort(pcb);
                return ERR_ABRT;
                
        } else {
                ++conn->retries;
                if (conn->retries == 64) {
                        owl_dbg("aborting pcb after retries:%d", conn->retries);
                        tcp_abort(pcb);
                        return ERR_ABRT;
                }

                send_data(pcb, conn, 1);
        }

        return ERR_OK;
}

/**
 * Return 1 if the connection have received a complete HTTP header.
 * We do some ugly string matching to find an empty line [CRLF] in the
 * received HTTP data.
 *
 */
static int
http_got_header(struct httpd_conn *conn)
{
        char buf[4] = { 0 };
        int offset;
        struct pbuf *p = conn->pbuf;

        owl_dbg("p->tot_len:%d", p->tot_len);
        
        for (offset = 0; offset < p->tot_len; offset++) {
                
                /* get the next 4 chars from our pbuf chain */
                if (pbuf_copy_partial(p, buf, 4, offset) < 4)
                        break;

                if (!owl_os_memcmp(buf, "\r\n\r\n", sizeof(buf))) {
                        owl_info("found http header at offset:%d", offset);
                        return 1;
                }
        }
        
        owl_info("http header not complete");
        owl_hexdump("CRLF", buf, sizeof(buf));

        return 0;
}


static void
recv_data(struct tcp_pcb *pcb, struct httpd_conn *conn,
          struct http_request *req)
{
        int len;
#if OWL_TRACE >= OWL_TRACE_DATA
        struct pbuf *iter = conn->pbuf;
        while (iter) {
                owl_hexdump("HTTP RX", iter->payload, iter->len);
                iter = iter->next;
        }
#endif
        
        len = conn->f(conn->ctx, 0, req, conn->pbuf,
                      conn->rspbuf, conn->rsplen, &conn->local);

        pbuf_free(conn->pbuf);
        conn->pbuf = NULL;
        
        /* A handler should wait for all expected data before sending a
           reply. */
        if (conn->left || conn->seq) {
                owl_err("handler received data during sending of reply, "
                        "conn->seq:%d conn->left:%d conn:0x%x",
                        conn->seq, conn->left, conn);
        }

        if (len > 0) {
                owl_info("handler:0x%x replied with %d bytes", conn->f, len);
                conn->left = len;
                conn->seq++;
                send_data(pcb, conn, 1);
                tcp_sent(pcb, http_sent);
                        
        } else if (len == OWL_ERR_AGAIN) {
                owl_dbg("handler:0x%x is waiting for more data", conn->f);

        } else {
                /* error or done from handler */
                owl_info("handler:0x%x returned %d, close conn", conn->f, len);
                close_conn(pcb, conn);
        }
}


/**
 * Find a document that matches the http request
 *
 */
static struct httpd_document *
find_document(struct list_head *documents, struct http_request *req,
              struct pbuf *p)
{
        struct list_head *pos = NULL;
        struct httpd_document *doc = NULL;

        owl_dbg("req:0x%x p:0x%x p->len:%d p->tot_len:%d",
                req, p, p->len, p->tot_len);
        
        list_for_each(pos, documents) {
                doc = list_entry(pos,
                                 struct httpd_document,
                                 list);
                
                owl_dbg("trying to match document with path:%s", doc->path);

#ifdef OWL_CONFIG_HTTPD_PARSER
                owl_unused(p);
                
                if (doc->method != req->method)
                        continue;

                if (!owl_os_strcmp(req->uri.path, "/") &&
                    !owl_os_strcmp(doc->path, "/index.html"))
                        return doc;
                
                if (!owl_os_strcmp(doc->path, req->uri.path))
                        return doc;
#else
                owl_unused(req);
                
                if (p->len < 5)  {
                        owl_info("header too short p->len:%d", p->len);
                        return NULL;
                }
                
                if (doc->method != HTTP_METHOD_GET) {
                        owl_info("doc has invalid method doc->method:%d",
                                 doc->method);
                        continue;
                }

                if (p->len < owl_os_strlen(doc->path) + 4) {
                        owl_info("header too short to fit document path "
                                 "p->len:%d doc->path:'%s'", p->len, doc->path);
                        continue;
                }
                
                if (!owl_os_strncmp(p->payload, "GET ", 4) &&
                    !owl_os_strncmp(doc->path, (char *) p->payload + 4,
                             owl_os_strlen(doc->path))) {
                        owl_dbg("found matching doc with path:%s", doc->path);
                        return doc;
                }

                if (!owl_os_strncmp(p->payload, "GET / ", 6) &&
                    !owl_os_strcmp(doc->path, "/index.html")) {
                        owl_dbg("found default doc");
                        return doc;
                }

#endif /* OWL_CONFIG_HTTPD_PARSER */
        }

        owl_info("no document found");
        return NULL;
}


/**
 * When data is received from the user agent, this function will be called.
 *
 * If the complete header has not yet been recieved, we will keep the pbuf and
 * wait for more data.
 *
 * As soon as this function is called with a pbuf that contains the final
 * part of the header, we will parse the header and try to match the method
 * and path with our list of documents.
 *
 * If no document is found, the 404 handler will be used.
 * If a document is found, but  authorization fails, the 401 handler will be
 * used. Otherwise, the document handler will be used to reply to the request.
 *
 */
static err_t
http_recv(void *arg, struct tcp_pcb *pcb, struct pbuf *p, err_t err)
{
        struct httpd_conn *conn = arg;
        struct httpd_document *doc = NULL;
        struct http_request *req = NULL;

        owl_unused(err);
        
        if (p == NULL) {
                owl_info("p was NULL, closing conn:0x%x", conn);
                close_conn(pcb, conn);
                return ERR_OK;
        }
        
        owl_dbg("conn:0x%x conn->pbuf:0x%x p:0x%x p->tot_len:%d",
                conn, conn->pbuf, p, p->tot_len);
        
        conn->retries = 0;

        /* build chain of pbufs */
	if (conn->pbuf == NULL)
		conn->pbuf = p;
	else
		pbuf_cat(conn->pbuf, p);
        
        /* Inform TCP that we have taken the data. */
        tcp_recved(pcb, p->tot_len);

        if (conn->f) {
                /* There is already a handler associated with this connection.
                 * The handler has already been invoked with the header,
                 * so invoke it again with more data.
                 *
                 * Set the seq to 0 and reqbuf to NULL in this call to
                 * indicate that this is more data coming from the original
                 * request.
                 */
                recv_data(pcb, conn, NULL);
                return ERR_OK;
                
        } else if (!http_got_header(conn)) {
                
                /* Check if we have received a complete header */
                owl_info("header not complete, waiting for more data");
                return ERR_OK;
        }

                                
#ifdef OWL_CONFIG_HTTPD_PARSER
        if ((req = httpd_parse_request(conn->pbuf)) == NULL) {
#if OWL_TRACE >= OWL_TRACE_DATA
                struct pbuf *iter = conn->pbuf;
                while (iter) {
                        owl_hexdump("HTTP", iter->payload, iter->len);
                        iter = iter->next;
                }
#endif /* OWL_TRACE */
                owl_err("could not parse http request, closing conn");
                
                close_conn(pcb, conn);
                return ERR_OK;
        }
#endif /* OWL_CONFIG_HTTPD_PARSER */

        
        if ((doc = find_document(&conn->httpd->documents, req,
                                 conn->pbuf)) == NULL) {
                /* 404 */
                owl_info("no document found, 404");
                conn->f = httpd_404;
                conn->ctx = NULL;
                
        } else {
#ifdef OWL_CONFIG_HTTPD_AUTH
                /* document found, check auth */
                struct http_auth *auth;

                if (doc->auth)
                        auth = doc->auth;
                else
                        auth = &conn->httpd->auth;
                
                owl_dbg("authorizating document with auth "
                        "scheme:%d user:%s passwd:%s realm:%s",
                        auth->scheme, auth->user, auth->password,
                        auth->realm);

                if (httpd_authorize(req, auth)) {
                        /* authorization failed, handle this with
                           the httpd_401 handler */

                        owl_info("authorization failed, 401");
                        conn->f = httpd_401;
                        conn->ctx = auth;
                        
                } else {

                        owl_dbg("authorization ok, set handler:0x%x", doc->f);
                        conn->f = doc->f;
                        conn->ctx = doc->ctx;
                }
#else
                owl_dbg("set handler:0x%x", doc->f);
                conn->f = doc->f;
                conn->ctx = doc->ctx;
#endif /* OWL_CONFIG_HTTPD_AUTH */
        }

                
        conn->seq = 0;
        recv_data(pcb, conn, req);
        /* pbuf free'd by recv_data */
#ifdef OWL_CONFIG_HTTPD_PARSER
        httpd_free_request(req);
#endif

        return ERR_OK;
}


static err_t
http_accept(void *arg, struct tcp_pcb *pcb, err_t err)
{
        struct httpd_ctx *priv = arg;
        struct httpd_conn *conn;
        
        owl_unused(err);

        tcp_setprio(pcb, TCP_PRIO_MIN);
  
        /* Allocate memory for the structure that holds the state of the
           connection. */
        if ((conn = owl_os_zalloc(sizeof(*conn))) == NULL) {
                owl_err("out of mem");
                return ERR_MEM;
        }

        conn->httpd = priv;
        conn->rsplen = OWL_CONFIG_HTTPD_RSPLEN;
        if ((conn->rspbuf = owl_os_alloc(conn->rsplen)) == NULL) {
                owl_os_free(conn);
                owl_err("out of mem");
                return ERR_MEM;
        }

        owl_info("conn:0x%x", conn);
        
        tcp_arg(pcb, conn);
        tcp_recv(pcb, http_recv);
        tcp_err(pcb, conn_err);
        tcp_poll(pcb, http_poll, 4);

        if (priv->pcb) {
                /* Defined as empty statement if unless TCP_LISTEN_BACKLOG=1 in
                   lwipopts.h */
                tcp_accepted(priv->pcb);
        }
        
        return ERR_OK;
}


/**
 * HTTPd connection manager
 *
 */
static void
conn_timer_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct httpd_ctx *priv = ctx;
        owl_unused(tmo);
        
        if (netif_is_up(owlnet_netif()) && priv->run && priv->pcb == NULL) {
                /* Start httpd if network is up and start was requested */
                
                err_t err;
                struct tcp_pcb *lpcb;
                
                priv->pcb = tcp_new();
                if (priv->pcb == NULL) {
                        owl_err("fail to alloc pcb");
                        return;
                }
                
                if ((err = tcp_bind(priv->pcb, IP_ADDR_ANY,
                                    priv->port)) != ERR_OK) {
                        owl_err("bind failed:%d", err);
                        tcp_abort(priv->pcb);
                        priv->pcb = NULL;
                        return;
                }
                
                if ((lpcb = tcp_listen(priv->pcb)) == NULL) {
                        owl_err("fail to listen");
                        tcp_abort(priv->pcb);
                        priv->pcb = NULL;
                        return;
                }
                
                /* priv->pcb was free'd by tcp_listen() */
                priv->pcb = lpcb;
                tcp_arg(priv->pcb, priv);
                tcp_accept(priv->pcb, http_accept);
        
        } else if ((!netif_is_up(owlnet_netif()) || !priv->run) &&
                   priv->pcb) {
                
                /* Stop httpd if network is down or stop requested */

                err_t err;
		err = tcp_close(priv->pcb);
                owl_assert(err == ERR_OK);
                priv->pcb = NULL;
        }
}

