/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_INFO

#include <lwip/tcp.h>

#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_net.h>

#include "ttcp.h"

struct ttcp_ctx {
        struct tcp_pcb *pcb;
        
        char *buf;
        uint16_t buflen; /* size of buffers */
        uint16_t nbuf;   /* number of buffers */
        uint16_t bufleft; /* remaining bytes to transfer in current buf */
        uint16_t bufcnt;  /* buffer index that is currently transmitted */ 
        uint32_t start_ms;
        uint32_t stop_ms;
        int send;

        uint32_t recvd; /* recv mode only */
        
        void (*progress_cb)(void *ctx, int progress);
        void (*done_cb)(void *ctx, int err, int bytes, uint32_t ms);
        void *ctx;
};

static void send_data(struct ttcp_ctx *priv);
static void tcp_err_cb(void *arg, err_t err);
static err_t tcp_connect_cb(void *arg, struct tcp_pcb *pcb, err_t err);
static err_t tcp_accept_cb(void *arg, struct tcp_pcb *pcb, err_t err);
static err_t tcp_recv_cb(void *arg, struct tcp_pcb *pcb, struct pbuf *p,
                         err_t err);
static int ttcp_done(struct ttcp_ctx *priv, int err);

/**
 * Start a ttcp connection for transmission. The ttcp transfer will
 * proceed asynchronously upon successful return.
 * 
 *
 * @return 0
 */
int
ttcp_connect(struct ip_addr *ip, uint16_t port, uint16_t nbuf,
             uint16_t buflen,
             void (*progress_cb)(void *ctx, int progress),
             void (*done_cb)(void *ctx, int err, int bytes, uint32_t ms),
             void *ctx)
{
        int err;
        struct ttcp_ctx *priv;
        
        owl_info("ip:%s port:%d nbuf:%d buflen:%d",
                 ip_ntoa(ip), port, nbuf, buflen);

        if (nbuf == 0 || buflen == 0 || port == 0)
                return OWL_ERR_PARAM;

        /* Allocate our private ttcp struct */
        if ((priv = owl_os_zalloc(sizeof(struct ttcp_ctx))) == NULL) {
                owl_info("fail to alloc priv");
                return OWL_ERR_MEM;
        }

        priv->nbuf = nbuf;
        priv->buflen = buflen;
        priv->bufleft = buflen;
        priv->progress_cb = progress_cb;
        priv->done_cb = done_cb;
        priv->ctx = ctx;
        priv->send = 1;
        
        /* Allocate a buffer to hold the data we're going to transmit */
        if ((priv->buf = owl_os_alloc(priv->buflen)) == NULL) {
                owl_info("fail to alloc buf (%d b)", priv->buflen);
                return ttcp_done(priv, OWL_ERR_MEM);
        }
        
        /* Create a TCP socket */
        if ((priv->pcb = tcp_new()) == NULL) {
                owl_err("fail to alloc tcp pcb");
                return ttcp_done(priv, OWL_ERR_MEM);
        }
        owl_dbg("tcp_new:%p", priv->pcb); 
        
        tcp_arg(priv->pcb, priv);
        tcp_err(priv->pcb, tcp_err_cb);
        tcp_recv(priv->pcb, tcp_recv_cb);

        owl_dbg("tcp_connect:%p", priv->pcb);
        if ((err = tcp_connect(priv->pcb, ip, port,
                               tcp_connect_cb)) != ERR_OK) {
                owl_err("fail to conn err:%d", err);
                return ttcp_done(priv, lwip_error(err));
        }

        /* Wait for tcp_connect_cb() or tcp_err_cb() */
        return 0;
}


/**
 * Start a ttcp connection for reception. The ttcp transfer will
 * proceed asyncrously upon successful return.
 * 
 *
 * @return sockid if started successfully.
 */
#if __GNUC__
#pragma GCC diagnostic ignored "-Wcast-qual" /* IP_ADDR_ANY */
#endif
int
ttcp_listen(uint16_t port,
            void (*progress_cb)(void *ctx, int progress),
            void (*done_cb)(void *ctx, int err, int bytes, uint32_t ms),
            void *ctx)
{
        int err;
        struct tcp_pcb *lpcb;
        struct ttcp_ctx *priv;
        
        owl_info("port:%d", port);

        /* Allocate our private ttcp struct */
        if ((priv = owl_os_zalloc(sizeof(struct ttcp_ctx))) == NULL) {
                owl_info("fail to alloc priv");
                return ttcp_done(priv, OWL_ERR_MEM);
        }

        priv->progress_cb = progress_cb;
        priv->done_cb = done_cb;
        priv->ctx = ctx;

        /* Create a TCP socket */
        if ((priv->pcb = tcp_new()) == NULL) {
                owl_err("fail to alloc tcp pcb");
                return ttcp_done(priv, OWL_ERR_MEM);
        }
        owl_dbg("tcp_new:%p", priv->pcb); 
        
        if ((err = tcp_bind(priv->pcb, IP_ADDR_ANY, port)) != ERR_OK) {
                owl_err("bind failed:%d", err);
                tcp_abort(priv->pcb);
                priv->pcb = NULL;
                return ttcp_done(priv, lwip_error(err));
        }
                
        if ((lpcb = tcp_listen(priv->pcb)) == NULL) {
                owl_err("fail to listen");
                tcp_abort(priv->pcb);
                priv->pcb = NULL;
                return ttcp_done(priv, OWL_ERR_MEM);
        }
                
        /* priv->pcb was free'd by tcp_listen() */
        priv->pcb = lpcb;
        tcp_arg(priv->pcb, priv);
        tcp_accept(priv->pcb, tcp_accept_cb);

        /* Wait for tcp_accept_cb() */
        return 0;
}
#if __GNUC__
#pragma GCC diagnostic warning "-Wcast-qual"
#endif

static void
send_data(struct ttcp_ctx *priv)
{
        err_t err;
        
        owl_dbg("bufcnt:%d bufleft:%d tcp_sndbuf:%d",
                priv->bufcnt, priv->bufleft, tcp_sndbuf(priv->pcb));

        /* NOTE: tcp_sent_cb will be called when we get an ack from the remote
           end. However, the remote end will not send an ack unless we filled
           up the TCP window. So for optimal throughput, the TCP_SND_BUF need to
           be as large as TCP_WND and we need to fill it up (so make sure that
           we use ttcp buffers that have the same size as TCP_SND_BUF */

        while (tcp_sndbuf(priv->pcb)) {

                int len = priv->bufleft;
        
                /* all sent */
                if (priv->bufcnt == priv->nbuf) {
                        ttcp_done(priv, 0);
                        return;
                }
                
                /* We cannot send more data than space available in the send
                   buffer. */
                if (len > tcp_sndbuf(priv->pcb))
                        len = tcp_sndbuf(priv->pcb);

                if ((err = tcp_write(priv->pcb, priv->buf, len, 0)) != ERR_OK) {
                        /* out of pbufs */
                        owl_info("fail to write len:%d err:%d", len, err);
                        return;
                }

                tcp_output(priv->pcb);
                owl_dbg("wrote %d bytes", len);

                priv->bufleft -= len;
                
                if (priv->bufleft == 0) {
                        if (priv->progress_cb)
                                priv->progress_cb(priv->ctx, priv->bufcnt);
                
                        priv->bufleft = priv->buflen;
                        priv->bufcnt++;
                }
        }
}


static err_t
tcp_sent_cb(void *arg, struct tcp_pcb *pcb, u16_t len)
{
        struct ttcp_ctx *priv = arg;

        (void) pcb;
        (void) len;
        
        send_data(priv);
        return ERR_OK;
}

static void 
tcp_err_cb(void *arg, err_t err)
{
        struct ttcp_ctx *priv = arg;
        (void) err;
        owl_info("err:%d", err);
        priv->pcb = NULL; /* free'd by lwip upon return */
        ttcp_done(priv, OWL_ERR_CONN);
}


static err_t 
tcp_connect_cb(void *arg, struct tcp_pcb *pcb, err_t err)
{
        struct ttcp_ctx *priv = arg;

        (void) pcb;
        (void) err;
        
        owl_assert(priv->send);
        owl_info("ttcp connected, start send");
        
        priv->start_ms = owltmr_get_ms();

        tcp_sent(pcb, tcp_sent_cb);
        send_data(priv);

        return ERR_OK;
}


static err_t
tcp_accept_cb(void *arg, struct tcp_pcb *pcb, err_t err)
{
        struct ttcp_ctx *priv = arg;
        (void) err;

        owl_dbg("pcb:%p", pcb);
        
        if (tcp_close(priv->pcb) != ERR_OK) {
                owl_err("fail to close listeing pcb, aborting pcb");
                tcp_abort(priv->pcb);
        }
        
        priv->pcb = pcb;
        tcp_arg(pcb, priv);
        tcp_recv(pcb, tcp_recv_cb);
        tcp_err(pcb, tcp_err_cb);

        priv->start_ms = owltmr_get_ms();
        return ERR_OK;
}


static err_t 
tcp_recv_cb(void *arg, struct tcp_pcb *pcb, struct pbuf *p, err_t err)
{
        struct ttcp_ctx* priv = arg;

        (void) err;

        owl_dbg("p:%p p->tot_len:%d", p, p ? p->tot_len : 0);

        if (priv->send) {
                owl_info("ignoring rx data");
                return ERR_OK;
        }
        
        /* p will be NULL when remote end is done */
        if (p == NULL) {
                ttcp_done(priv, 0);
                return ERR_OK;
        }

        priv->recvd += p->tot_len;
        priv->bufcnt++;

        if (priv->progress_cb)
                priv->progress_cb(priv->ctx, priv->bufcnt);
        
        tcp_recved(pcb, p->tot_len);
        pbuf_free(p);
        return ERR_OK;
}

static int
ttcp_done(struct ttcp_ctx *priv, int err)
{
        uint32_t bytes, ms;

        owl_info("err:%d", err);

        priv->stop_ms = owltmr_get_ms();
        
        ms = owltmr_get_intvl(priv->start_ms, priv->stop_ms);
        if (priv->send)
                bytes = priv->nbuf * priv->buflen;
        else
                bytes = priv->recvd;

        if (priv->done_cb)
                priv->done_cb(priv->ctx, err, bytes, ms);

        if (priv->pcb) {
                tcp_arg(priv->pcb, NULL);
                tcp_sent(priv->pcb, NULL);
                tcp_recv(priv->pcb, NULL);
                tcp_close(priv->pcb);
        }

        owl_os_free(priv->buf);
        owl_os_free(priv);
        return err;
}


static void
progress_cb(void *ctx, int i)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, ".");
        if (i > 0 && i % 80 == 0)
                owlsh_printf(sh, "\n\r");
}

static void
done_cb(void *ctx, int err, int bytes, uint32_t ms)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "\n\r");
        
        if (err)
                owlsh_printf(sh, "ttcp err:%d (%s)\n\r", err, owl_error(err));
                
        else if (ms > 0)
                owlsh_printf(sh, "ttcp ok: %d bytes processed in %lu ms, "
                             "%lu.%lu KB/s\n\r",
                             bytes, ms, bytes / ms, bytes % ms);
        else
                owlsh_printf(sh, "ttcp ok: %d bytes processed in 0 ms\n\r",
                             bytes);
}

static void
ttcp_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int i;
        uint16_t port = 5000;          /* -p */
        uint16_t nbuf = 1024;          /* -n */
        uint16_t buflen = 1024;        /* -l */
        int connect = -1;              /* -t , -r */
        struct ip_addr ip = ip_addr_any;
        int err;

        void (*progress)(void *ctx, int progress) = progress_cb;
        
	/* parse arguments, ugly */
	for (i = 1; i < argc; i++) {
                
                if (owl_os_strcmp(argv[i], "-p") == 0) {
                        /* -p <port> */
			if (++i >= argc)
				goto err_parse;
                        
                        port = owl_os_atoi(argv[i]);
                        
                } else if (owl_os_strcmp(argv[i], "-n") == 0) {
                        
                        /* -n <nbuf> */
			if (++i >= argc)
				goto err_parse;
                        
                        nbuf = owl_os_atoi(argv[i]);

		} else if (owl_os_strcmp(argv[i], "-l") == 0) {
			/* -l <buflen> */
			if (++i >= argc)
				goto err_parse;
                        
                        buflen = owl_os_atoi(argv[i]);

		} else if (owl_os_strcmp(argv[i], "-t") == 0) {
                        /* transmit */
                        connect = 1;

                } else if (owl_os_strcmp(argv[i], "-r") == 0) {
                        /* receieve */
                        connect = 0;

                } else if (owl_os_strcmp(argv[i], "-q") == 0) {
                        /* no progress cb that will print stuff */
                        progress = NULL;

		} else if (!ip_aton(argv[i], &ip)) {
                        /* ip address */
                        goto err_parse;
                }
        }

        if (connect < 0 || (connect && ip_addr_cmp(&ip, &ip_addr_any)))
                goto err_parse;
        
        if (connect)
                err = ttcp_connect(&ip, port, nbuf, buflen,
                                   progress, done_cb, sh);
        else
                err = ttcp_listen(port, progress, done_cb, sh);

        if (err < 0)
                owlsh_printf(sh, "%s %d (%s)\n\r", err >= 0 ? "ok" : "err", err,
                             owl_error(err));
        else if (connect)
                owlsh_printf(sh, "ttcp connect %s:%d nbuf:%d buflen:%d\n\r",
                             ip_ntoa(&ip), port, nbuf, buflen);
        else 
                owlsh_printf(sh, "ttcp listen port:%d\n\r", port);

        return;
        
err_parse:
        owlsh_printf(sh, "usage: ttcp [-trpnlq]\n\r");
        owlsh_printf(sh, "options:\n\r");
        owlsh_printf(sh, "  -t         transmit\n\r");
        owlsh_printf(sh, "  -r         receive\n\r");
        owlsh_printf(sh, "  -p <port>  port\n\r");
        owlsh_printf(sh, "  -n <count> number of buffers (transmit only)\n\r");
        owlsh_printf(sh, "  -l <size>  size of buffers (trasmit only)\n\r");
        owlsh_printf(sh, "  -q         silent\n\r");
        
}

const struct owlsh_cmd cmd_ttcp = {
        "ttcp", ttcp_f, "ttcp throughput test"
};

